from cone.app import security
from cone.app import testing
from cone.app.interfaces import IACLAdapter
from cone.app.interfaces import IAdapterACL
from cone.app.interfaces import IApplicationNode
from cone.app.interfaces import IAuthenticator
from cone.app.interfaces import IOwnerSupport
from cone.app.interfaces import IPrincipalACL
from cone.app.model import BaseNode
from cone.app.security import acl_registry
from cone.app.security import AdapterACL
from cone.app.security import authenticate
from cone.app.security import authenticated_user
from cone.app.security import DEFAULT_ACL
from cone.app.security import groups_callback
from cone.app.security import logger
from cone.app.security import OwnerSupport
from cone.app.security import principal_by_id
from cone.app.security import PrincipalACL
from cone.app.security import search_for_principals
from cone.app.ugm import ugm_backend
from node.ext.ugm.interfaces import IGroup
from node.ext.ugm.interfaces import IUser
from node.tests import NodeTestCase
from node.utils import instance_property
from plumber import default
from plumber import plumbing
from pyramid.authentication import AuthTktAuthenticationPolicy
from pyramid.interfaces import IAuthenticationPolicy
from pyramid.security import ACLAllowed
from pyramid.security import ACLDenied
from pyramid.security import ALL_PERMISSIONS
from pyramid.threadlocal import get_current_registry
from zope.component import adapter
from zope.component.globalregistry import BaseGlobalComponents
from zope.interface import implementer
import logging


@implementer(IAuthenticator)
class TestAuthenticator(object):

    def authenticate(self, login, password):
        return login


class SecurityTest(NodeTestCase):
    layer = testing.security

    def test_superuser_credentials(self):
        # Superuser credentials are set on application startup by main function
        # in cone.app.__init__

        # 'superuser' values get set in test layer
        self.assertEqual(security.ADMIN_USER, 'superuser')
        self.assertEqual(security.ADMIN_PASSWORD, 'superuser')

    def test_authenticated_user(self):
        self.layer.new_request()
        self.assertTrue(authenticated_user(self.layer.current_request) is None)

        with self.layer.authenticated('manager'):
            user = authenticated_user(self.layer.current_request)
            self.assertTrue(IUser.providedBy(user))
            self.assertEqual(user.name, 'manager')

    def test_principal_by_id(self):
        user = principal_by_id('manager')
        self.assertTrue(IUser.providedBy(user))
        self.assertEqual(user.name, 'manager')

        group = principal_by_id('group:group1')
        self.assertTrue(IGroup.providedBy(group))
        self.assertEqual(group.name, 'group1')

        self.assertTrue(principal_by_id('inexistent') is None)

    def test_search_for_principals(self):
        self.assertEqual(search_for_principals('viewer'), [u'viewer'])
        self.assertEqual(search_for_principals('group*'), [u'group:group1'])

    def test_acls(self):
        # The default ACL
        self.assertEqual(security.DEFAULT_ACL, [
            ('Allow', 'system.Authenticated', ['view']),
            ('Allow', 'role:viewer', ['view', 'list']),
            ('Allow', 'role:editor', [
                'view', 'list', 'add', 'edit', 'change_order'
            ]),
            ('Allow', 'role:admin', [
                'view', 'list', 'add', 'edit', 'change_order', 'delete', 'cut',
                'copy', 'paste', 'manage_permissions', 'change_state'
            ]),
            ('Allow', 'role:manager', [
                'view', 'list', 'add', 'edit', 'change_order', 'delete', 'cut',
                'copy', 'paste', 'manage_permissions', 'change_state', 'manage'
            ]),
            ('Allow', 'role:owner', [
                'view', 'list', 'add', 'edit', 'change_order', 'delete', 'cut',
                'copy', 'paste', 'manage_permissions', 'change_state'
            ]),
            ('Allow', 'system.Everyone', ['login']),
            ('Deny', 'system.Everyone', ALL_PERMISSIONS)
        ])

        # Base security tests
        policy = get_current_registry().queryUtility(IAuthenticationPolicy)
        self.assertTrue(isinstance(policy, AuthTktAuthenticationPolicy))

        self.layer.new_request()
        self.assertTrue(isinstance(
            self.layer.current_request.registry,
            BaseGlobalComponents
        ))

        with self.layer.authenticated('inexistent'):
            userid = self.layer.current_request.authenticated_userid
            self.assertTrue(userid is None)

        # Create some security context for testing
        class ACLTest(object):
            __acl__ = DEFAULT_ACL
        context = ACLTest()

        # Authenticate several users and check permission
        with self.layer.authenticated('superuser', 'superuser'):
            userid = self.layer.current_request.authenticated_userid
            self.assertEqual(userid, 'superuser')

            rule = self.layer.current_request.has_permission('manage', context)
            self.assertTrue(isinstance(rule, ACLAllowed))

        with self.layer.authenticated('viewer'):
            userid = self.layer.current_request.authenticated_userid
            self.assertEqual(userid, 'viewer')

            rule = self.layer.current_request.has_permission('manage', context)
            self.assertTrue(isinstance(rule, ACLDenied))

        userid = self.layer.current_request.authenticated_userid
        self.assertTrue(userid is None)

        rule = self.layer.current_request.has_permission('manage', context)
        self.assertTrue(isinstance(rule, ACLDenied))

    def test_ACLRegistry(self):
        class SomeModel(object):
            pass

        acl = [('Allow', 'role:viewer', ['view'])]
        acl_registry.register(acl, SomeModel)

        acl = [('Allow', 'role:viewer', ['edit'])]
        acl_registry.register(acl, node_info_name='some_model')

        acl = [('Allow', 'role:viewer', ['delete'])]
        acl_registry.register(acl, SomeModel, 'some_model')

        self.assertEqual(
            acl_registry.lookup(None, None, [('Allow', 'role:viewer', ['add'])]),
            [('Allow', 'role:viewer', ['add'])]
        )
        self.assertEqual(
            acl_registry.lookup(SomeModel),
            [('Allow', 'role:viewer', ['view'])]
        )
        self.assertEqual(
            acl_registry.lookup(node_info_name='some_model'),
            [('Allow', 'role:viewer', ['edit'])]
        )
        self.assertEqual(
            acl_registry.lookup(SomeModel, 'some_model'),
            [('Allow', 'role:viewer', ['delete'])]
        )

    def test_OwnerSupport(self):
        @plumbing(OwnerSupport)
        class OwnerSupportNode(BaseNode):
            pass

        ownersupportnode = OwnerSupportNode()
        self.assertTrue(IOwnerSupport.providedBy(ownersupportnode))
        self.assertTrue(ownersupportnode.owner is None)
        self.assertEqual(
            ownersupportnode.__acl__[0],
            ('Allow', 'system.Authenticated', ['view'])
        )

        with self.layer.authenticated('sepp'):
            userid = self.layer.current_request.authenticated_userid
            self.assertEqual(userid, 'sepp')

            ownersupportnode = OwnerSupportNode()
            self.assertEqual(ownersupportnode.owner, 'sepp')
            self.assertEqual(ownersupportnode.attrs['owner'], 'sepp')
            self.assertEqual(ownersupportnode.__acl__, [
                ('Allow', 'sepp', [
                    'view', 'list', 'add', 'edit', 'change_order', 'delete',
                    'cut', 'copy', 'paste', 'manage_permissions', 'change_state'
                ]),
                ('Allow', 'system.Authenticated', ['view']),
                ('Allow', 'role:viewer', ['view', 'list']),
                ('Allow', 'role:editor', [
                    'view', 'list', 'add', 'edit', 'change_order'
                ]),
                ('Allow', 'role:admin', [
                    'view', 'list', 'add', 'edit', 'change_order', 'delete',
                    'cut', 'copy', 'paste', 'manage_permissions',
                    'change_state'
                ]),
                ('Allow', 'role:manager', [
                    'view', 'list', 'add', 'edit', 'change_order', 'delete',
                    'cut', 'copy', 'paste', 'manage_permissions',
                    'change_state', 'manage'
                ]),
                ('Allow', 'role:owner', [
                    'view', 'list', 'add', 'edit', 'change_order', 'delete',
                    'cut', 'copy', 'paste', 'manage_permissions',
                    'change_state'
                ]),
                ('Allow', 'system.Everyone', ['login']),
                ('Deny', 'system.Everyone', ALL_PERMISSIONS)
            ])

        with self.layer.authenticated('viewer'):
            rule = self.layer.current_request.has_permission(
                'delete',
                ownersupportnode
            )
            self.assertTrue(isinstance(rule, ACLDenied))

        with self.layer.authenticated('sepp'):
            rule = self.layer.current_request.has_permission(
                'delete',
                ownersupportnode
            )
            self.assertTrue(isinstance(rule, ACLAllowed))

        @plumbing(OwnerSupport)
        class NoOwnerACLOnBaseNode(BaseNode):
            @property
            def __acl__(self):
                return [('Allow', 'role:viewer', ['view'])]

        with self.layer.authenticated('sepp'):
            ownersupportnode = NoOwnerACLOnBaseNode()
            self.assertEqual(ownersupportnode.owner, 'sepp')
            self.assertEqual(
                ownersupportnode.__acl__,
                [('Allow', 'role:viewer', ['view'])]
            )

        OwnerSupportNode.owner_attribute_name = 'userid'
        with self.layer.authenticated('sepp'):
            node = OwnerSupportNode()
        self.assertEqual(node.attrs['userid'], 'sepp')

    def test_PrincipalACL(self):
        # PrincipalACL is an abstract class. Directly use causes an error
        @plumbing(PrincipalACL)
        class PrincipalACLNode(BaseNode):
            pass

        node = PrincipalACLNode()
        err = self.expect_error(NotImplementedError, lambda: node.__acl__)
        expected = (
            'Abstract ``PrincipalACL`` does not implement ``principal_roles``.'
        )
        self.assertEqual(str(err), expected)

        # Concrete PrincipalACL implementation. Implements principal_roles
        # property
        class MyPrincipalACL(PrincipalACL):
            @default
            @instance_property
            def principal_roles(self):
                return dict()

        @plumbing(MyPrincipalACL)
        class MyPrincipalACLNode(BaseNode):
            pass

        node = MyPrincipalACLNode()
        self.assertTrue(IPrincipalACL.providedBy(node))

        node.principal_roles['someuser'] = ['manager']
        node.principal_roles['otheruser'] = ['editor']
        node.principal_roles['group:some_group'] = ['editor', 'manager']

        def find_rule(acl, who):
            for rule in acl:
                if rule[1] == who:
                    return rule

        rule = find_rule(node.__acl__, 'someuser')
        self.assertEqual(rule[0], 'Allow')
        self.assertEqual(rule[1], 'someuser')
        self.assertEqual(sorted(rule[2]), sorted([
            'cut', 'change_order', 'edit', 'copy', 'manage', 'list', 'add',
            'change_state', 'view', 'paste', 'manage_permissions', 'delete'
        ]))

        rule = find_rule(node.__acl__, 'otheruser')
        self.assertEqual(rule[0], 'Allow')
        self.assertEqual(rule[1], 'otheruser')
        self.assertEqual(sorted(rule[2]), sorted([
            'edit', 'add', 'list', 'view', 'change_order'
        ]))

        rule = find_rule(node.__acl__, 'group:some_group')
        self.assertEqual(rule[0], 'Allow')
        self.assertEqual(rule[1], 'group:some_group')
        self.assertEqual(sorted(rule[2]), sorted([
            'cut', 'edit', 'copy', 'manage', 'list', 'add', 'change_state',
            'view', 'paste', 'manage_permissions', 'delete', 'change_order'
        ]))

        rule = find_rule(node.__acl__, 'system.Authenticated')
        self.assertEqual(rule[0], 'Allow')
        self.assertEqual(rule[1], 'system.Authenticated')
        self.assertEqual(rule[2], ['view'])

        rule = find_rule(node.__acl__, 'role:viewer')
        self.assertEqual(rule[0], 'Allow')
        self.assertEqual(rule[1], 'role:viewer')
        self.assertEqual(sorted(rule[2]), sorted(['view', 'list']))

        rule = node.__acl__[-1]
        self.assertEqual(rule[0], 'Deny')
        self.assertEqual(rule[1], 'system.Everyone')
        self.assertEqual(rule[2], ALL_PERMISSIONS)

        # PrincipalACL role inheritance
        child = node['child'] = MyPrincipalACLNode()
        child.principal_roles['someuser'] = ['editor']

        rule = find_rule(child.__acl__, 'someuser')
        self.assertEqual(rule[0], 'Allow')
        self.assertEqual(rule[1], 'someuser')
        self.assertEqual(sorted(rule[2]), sorted([
            'edit', 'add', 'list', 'view', 'change_order'
        ]))

        rule = find_rule(child.__acl__, 'system.Authenticated')
        self.assertEqual(rule[0], 'Allow')
        self.assertEqual(rule[1], 'system.Authenticated')
        self.assertEqual(rule[2], ['view'])

        rule = find_rule(child.__acl__, 'role:viewer')
        self.assertEqual(rule[0], 'Allow')
        self.assertEqual(rule[1], 'role:viewer')
        self.assertEqual(sorted(rule[2]), sorted(['view', 'list']))

        rule = child.__acl__[-1]
        self.assertEqual(rule[0], 'Deny')
        self.assertEqual(rule[1], 'system.Everyone')
        self.assertEqual(rule[2], ALL_PERMISSIONS)

        subchild = child['child'] = MyPrincipalACLNode()
        subchild.role_inheritance = True
        subchild.principal_roles['otheruser'] = ['admin']
        self.assertEqual(subchild.aggregated_roles_for('inexistent'), [])
        self.assertEqual(
            sorted(subchild.aggregated_roles_for('someuser')),
            sorted(['manager', 'editor'])
        )
        self.assertEqual(
            sorted(subchild.aggregated_roles_for('otheruser')),
            sorted(['admin', 'editor'])
        )

        rule = find_rule(subchild.__acl__, 'someuser')
        self.assertEqual(rule[0], 'Allow')
        self.assertEqual(rule[1], 'someuser')
        self.assertEqual(sorted(rule[2]), sorted([
            'cut', 'edit', 'copy', 'manage', 'list', 'add', 'change_state',
            'view', 'paste', 'manage_permissions', 'delete', 'change_order'
        ]))

        rule = find_rule(subchild.__acl__, 'otheruser')
        self.assertEqual(rule[0], 'Allow')
        self.assertEqual(rule[1], 'otheruser')
        self.assertEqual(sorted(rule[2]), sorted([
            'cut', 'edit', 'copy', 'list', 'add', 'change_state', 'view',
            'paste', 'manage_permissions', 'delete', 'change_order'
        ]))

        rule = find_rule(subchild.__acl__, 'group:some_group')
        self.assertEqual(rule[0], 'Allow')
        self.assertEqual(rule[1], 'group:some_group')
        self.assertEqual(sorted(rule[2]), sorted([
            'cut', 'edit', 'copy', 'manage', 'list', 'add', 'change_state',
            'view', 'paste', 'manage_permissions', 'delete', 'change_order'
        ]))

        rule = find_rule(subchild.__acl__, 'system.Authenticated')
        self.assertEqual(rule[0], 'Allow')
        self.assertEqual(rule[1], 'system.Authenticated')
        self.assertEqual(rule[2], ['view'])

        rule = subchild.__acl__[-1]
        self.assertEqual(rule[0], 'Deny')
        self.assertEqual(rule[1], 'system.Everyone')
        self.assertEqual(rule[2], ALL_PERMISSIONS)

        # Principal roles get inherited even if some parent does not provide
        # principal roles
        child = node['no_principal_roles'] = BaseNode()
        subchild = child['no_principal_roles'] = MyPrincipalACLNode()
        self.assertEqual(
            sorted(subchild.aggregated_roles_for('group:some_group')),
            sorted(['manager', 'editor'])
        )

        # If principal role found which is not provided by plumbing endpoint
        # acl, this role does not grant any permissions
        node = MyPrincipalACLNode()
        node.principal_roles['someuser'] = ['inexistent_role']

        rule = find_rule(node.__acl__, 'someuser')
        self.assertEqual(rule[0], 'Allow')
        self.assertEqual(rule[1], 'someuser')
        self.assertEqual(rule[2], [])

        rule = find_rule(node.__acl__, 'system.Authenticated')
        self.assertEqual(rule[0], 'Allow')
        self.assertEqual(rule[1], 'system.Authenticated')
        self.assertEqual(rule[2], ['view'])

        rule = find_rule(node.__acl__, 'role:viewer')
        self.assertEqual(rule[0], 'Allow')
        self.assertEqual(rule[1], 'role:viewer')
        self.assertEqual(sorted(rule[2]), sorted(['view', 'list']))

        rule = node.__acl__[-1]
        self.assertEqual(rule[0], 'Deny')
        self.assertEqual(rule[1], 'system.Everyone')
        self.assertEqual(rule[2], ALL_PERMISSIONS)
        self.assertEqual(
            node.__acl__[-1],
            ('Deny', 'system.Everyone', ALL_PERMISSIONS)
        )

    def test_AdapterACL(self):
        @plumbing(AdapterACL)
        class AdapterACLNode(BaseNode):
            pass

        node = AdapterACLNode()
        self.assertEqual(node.__acl__, security.DEFAULT_ACL)
        self.assertTrue(IAdapterACL.providedBy(node))

        acl = [('Allow', 'role:viewer', ['view'])]

        @implementer(IACLAdapter)
        @adapter(IApplicationNode)
        class ACLAdapter(object):
            def __init__(self, model):
                self.model = model

            @property
            def acl(self):
                return acl

        request = self.layer.new_request()
        request.registry.registerAdapter(ACLAdapter)

        acl_adapter = request.registry.queryAdapter(
            node,
            IACLAdapter,
            default=None
        )
        self.assertIsInstance(acl_adapter, ACLAdapter)
        self.assertEqual(acl_adapter.acl, acl)
        self.assertEqual(node.__acl__, acl)

        request.registry.unregisterAdapter(ACLAdapter)

    def test_authentication_logging(self):
        # If an authentication plugin raises an error when calling
        # ``authenticate``, an error message is logged
        class TestHandler(logging.StreamHandler):
            record = None

            def handle(self, record):
                self.record = record

        handler = TestHandler()

        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)

        orgin_ugm = ugm_backend.ugm
        ugm_backend.ugm = object()

        authenticate(self.layer.new_request(), 'foo', 'foo')

        self.check_output("""
        <LogRecord: cone.app, ..., ...security.py, ...,
        "Authentication plugin <... 'object'> raised an Exception while trying
        to authenticate: 'object' object has no attribute 'users'">
        """, str(handler.record))

        # Test Group callback, also logs if an error occurs
        with self.layer.authenticated('superuser', 'superuser'):
            self.assertEqual(
                groups_callback('superuser', self.layer.new_request()),
                [u'role:manager']
            )

        groups_callback('foo', self.layer.new_request())
        self.check_output("""
        <LogRecord: cone.app, 40,
        ...security.py, ..., "'object' object has no attribute 'users'">
        """, str(handler.record))

        # Cleanup
        logger.setLevel(logging.INFO)
        logger.removeHandler(handler)
        ugm_backend.ugm = orgin_ugm

    def test_IAuthenticator(self):
        request = self.layer.new_request()
        registry = request.registry
        authenticator = TestAuthenticator()
        registry.registerUtility(
            authenticator,
            IAuthenticator,
            name='test_authenticator'
        )
        authenticator_origin = security.AUTHENTICATOR
        security.AUTHENTICATOR = 'test_authenticator'

        with self.layer.authenticated('foo'):
            self.assertEqual(request.authenticated_userid, 'foo')
        self.assertEqual(request.authenticated_userid, None)

        security.AUTHENTICATOR = authenticator_origin
        registry.unregisterUtility(authenticator)



from cone.app.browser import get_related_view
from cone.app.browser import RelatedViewProvider
from cone.app.browser import render_main_template
from cone.app.browser.actions import ActionDelete
from cone.app.browser.actions import ActionEdit
from cone.app.browser.actions import ActionMoveDown
from cone.app.browser.actions import ActionMoveUp
from cone.app.browser.actions import ActionView
from cone.app.browser.actions import Toolbar
from cone.app.browser.actions import ViewLink
from cone.app.browser.copysupport import extract_copysupport_cookie
from cone.app.browser.table import RowData
from cone.app.browser.table import Table
from cone.app.browser.utils import make_query
from cone.app.browser.utils import make_url
from cone.app.interfaces import IApplicationNode
from cone.app.interfaces import ICopySupport
from cone.app.interfaces import IWorkflowState
from cone.tile import Tile
from cone.tile import tile
from node.interfaces import ILeaf
from node.utils import instance_property
from plumber import plumbing
from pyramid.i18n import TranslationStringFactory
from pyramid.view import view_config
import datetime


_ = TranslationStringFactory('cone.app')


FAR_PAST = datetime.datetime(2000, 1, 1)


class ContentsActionView(ActionView):
    """View action for contents table.

    Gets displayed in actions column.
    """
    title = ActionView.text
    text = None
    action = None
    event = 'contextchanged:#layout'


class ContentsActionEdit(ActionEdit):
    """Edit action for contents table.

    Gets displayed in actions column.
    """
    title = ActionEdit.text
    text = None
    action = None
    event = 'contextchanged:#layout'

    @property
    def target(self):
        query = make_query(contenttile='edit')
        return make_url(self.request, node=self.model, query=query)


class ContentsActionDelete(ActionDelete):
    """Delete action for contents table.

    Gets displayed in actions column.
    """
    title = ActionDelete.text
    text = None

    @property
    def display(self):
        return self.model.properties.action_delete \
            and self.request.has_permission('delete', self.model.parent) \
            and self.permitted('delete')


class ContentsActionMoveUp(ActionMoveUp):
    """Move up action for contents table.

    Gets displayed in actions column.
    """
    title = ActionMoveUp.text
    text = None


class ContentsActionMoveDown(ActionMoveDown):
    """Move down action for contents table.

    Gets displayed in actions column.
    """
    title = ActionMoveDown.text
    text = None


class ContentsViewLink(ViewLink):
    """View link for contents table.

    Title column uses this to turn title into view link.
    """
    css = 'title'
    event = 'contextchanged:#layout'
    action = None

    @property
    def target(self):
        related_view = get_related_view(self.request)
        if related_view and not ILeaf.providedBy(self.model):
            query = make_query(contenttile=related_view)
            return make_url(self.request, node=self.model, query=query)
        return super(ContentsViewLink, self).target


@tile(name='contents', path='templates/table.pt', permission='list')
class ContentsTile(Table):
    table_id = 'contents'
    table_tile_name = 'contents'
    col_defs = [{
        'id': 'actions',
        'title': _('actions', default='Actions'),
        'sort_key': None,
        'sort_title': None,
        'content': 'structure'
    }, {
        'id': 'title',
        'title': _('title', default='Title'),
        'sort_key': 'title',
        'sort_title': _('sort_on_title', default='Sort on title'),
        'content': 'structure'
    }, {
        'id': 'creator',
        'title': _('creator', default='Creator'),
        'sort_key': 'creator',
        'sort_title': _('sort_on_creator', default='Sort on creator'),
        'content': 'string'
    }, {
        'id': 'created',
        'title': _('created', default='Created'),
        'sort_key': 'created',
        'sort_title': _('sort_on_created', default='Sort on created'),
        'content': 'datetime'
    }, {
        'id': 'modified',
        'title': _('modified', default='Modified'),
        'sort_key': 'modified',
        'sort_title': _('sort_on_modified', default='Sort on modified'),
        'content': 'datetime'
    }]
    default_sort = 'created'
    default_order = 'desc'
    sort_keys = {
        'title': lambda x: x.metadata.title.lower(),
        'creator': lambda x: x.metadata.creator.lower(),
        'created': lambda x: (
            x.metadata.created and x.metadata.created or FAR_PAST
        ),
        'modified': lambda x: (
            x.metadata.modified and x.metadata.modified or FAR_PAST
        )
    }
    show_filter = True

    @property
    def item_count(self):
        return len(self.filtered_children)

    @instance_property
    def row_actions(self):
        row_actions = Toolbar()
        row_actions['up'] = ContentsActionMoveUp()
        row_actions['down'] = ContentsActionMoveDown()
        row_actions['view'] = ContentsActionView()
        row_actions['edit'] = ContentsActionEdit()
        row_actions['delete'] = ContentsActionDelete()
        return row_actions

    @instance_property
    def view_link(self):
        return ContentsViewLink()

    def row_data(self, node):
        row_data = RowData()
        row_data['actions'] = self.row_actions(node, self.request)
        row_data['title'] = self.view_link(node, self.request)
        row_data['creator'] = node.metadata.get('creator', 'unknown')
        row_data['created'] = node.metadata.get('created')
        row_data['modified'] = node.metadata.get('modified')
        return row_data

    def sorted_rows(self, start, end, sort, order):
        children = self.sorted_children(sort, order)
        rows = list()
        cut_urls = extract_copysupport_cookie(self.request, 'cut')
        for child in children[start:end]:
            row_data = self.row_data(child)
            target = make_url(self.request, node=child)
            if ICopySupport.providedBy(child):
                row_data.selectable = True
                row_data.target = target
                row_data.css = 'copysupportitem'
                if target in cut_urls:
                    row_data.css += ' copysupport_cut'
            if IWorkflowState.providedBy(child):
                row_data.css += ' state-%s' % child.state
            if hasattr(child, 'node_info_name') and child.node_info_name:
                row_data.css += ' node-type-%s' % child.node_info_name
            rows.append(row_data)
        return rows

    @property
    def listable_children(self):
        return [
            child for child in self.model.values()
            if IApplicationNode.providedBy(child)
        ]

    @property
    def filtered_children(self):
        if '_filtered_children' in self.request.environ:
            return self.request.environ['_filtered_children']
        children = list()
        term = self.filter_term
        if term:
            term = term.lower()
        for node in self.listable_children:
            if not self.request.has_permission('view', node):
                continue
            if term:
                metadata = node.metadata
                title = metadata.get('title')
                title = title.lower() if title else ''
                creator = metadata.get('creator')
                creator = creator.lower() if creator else ''
                if title.find(term) == -1 and creator.find(term) == -1:
                    continue
            children.append(node)
        self.request.environ['_filtered_children'] = children
        return children

    def sorted_children(self, sort, order):
        children = self.filtered_children
        if sort in self.sort_keys:
            children = sorted(children, key=self.sort_keys[sort])
            if order == 'asc':
                children.reverse()
        return children


@tile(name='listing', path='templates/listing.pt', permission='list')
@plumbing(RelatedViewProvider)
class ListingTile(Tile):
    """Tile rendering the listing.
    """
    related_view = 'listing'


@view_config(name='listing', permission='list')
def listing(model, request):
    """Listing view.
    """
    return render_main_template(model, request, 'listing')



from node.interfaces import IAttributes
from node.interfaces import IChildFactory
from node.interfaces import ILeaf
from node.interfaces import INode
from node.interfaces import ISchema
from node.interfaces import IUUIDAware
from zope.interface import Attribute
from zope.interface import Interface
from zope.interface.common.mapping import IReadMapping


class ISecured(Interface):
    """Secured object."""
    __acl__ = Attribute(u'ACL')


class IApplicationEnvironment(Interface):
    """Application environment."""
    request = Attribute(u'The current request if any.')
    registry = Attribute(u'The current registry.')


class IApplicationNode(ISecured, INode, IAttributes):
    """Application Node interface."""
    properties = Attribute(u'cone.app.interfaces.IProperties providing object')
    metadata = Attribute(u'cone.app.interfaces.IMetadata implementation')
    nodeinfo = Attribute(u'cone.app.interfaces.INodeInfo providing object')


class ILeafNode(IApplicationNode, ILeaf):
    """Application node with no children."""

    def __getitem__(name):
        """Raise ``KeyError``"""

    def __setitem__(name, value):
        """Raise ``KeyError``"""

    def __delitem__(name):
        """Raise ``KeyError``"""

    def __iter__():
        """Empty iterator."""


class IFactoryNode(IApplicationNode, IChildFactory):
    """Application node for static children."""


class IAdapterNode(IApplicationNode):
    """Application node which acts as adapter on any context.

    XXX: - currently designed to adapt nodes, more generic
         - no attrs on this interface
         - self.context instead of self.model
    """
    attrs = Attribute(u'Return self.model.attrs')

    def __init__(model, name, parent):
        """Name and parent are used to hook the correct application hierarchy.
        """


class IProperties(IReadMapping):
    """Interface for providing any kind of properties."""

    def __getattr__(name):
        """Return property by attribute access.

        Never throws an AttributeError if attribute does not exists, return
        None instead.
        """

    def __setattr__(name, value):
        """Set property by attribute access.
        """

    def keys():
        """Return available properties
        """


class ILayoutConfig(IProperties):
    """Layout configuration."""
    mainmenu = Attribute(u'Flag whether to display mainmenu')
    mainmenu_fluid = Attribute(u'Flag whether mainmenu is fluid')
    livesearch = Attribute(u'Flag whether to display livesearch')
    personaltools = Attribute(u'Flag whether to display personaltools')
    columns_fluid = Attribute(u'Flag whether columns are fluid')
    pathbar = Attribute(u'Flag whether to display pathbar')
    sidebar_left = Attribute(u'Tiles which should be rendered in sidebar')
    sidebar_left_grid_width = Attribute(u'Sidebar grid width')
    content_grid_width = Attribute(u'Content grid width')


# B/C, removed as of cone.app 1.1
ILayout = ILayoutConfig


class IMetadata(IProperties):
    """Interface for providing metadata for application nodes."""


class INodeInfo(IProperties):
    """Interface for providing node information."""
    title = Attribute(u'Node meta title.')
    description = Attribute(u'Node meta description.')
    node = Attribute(u'Node implementing class.')
    factory = Attribute(u'Add model factory.')
    addables = Attribute(u'List of valid children node info names.')
    icon = Attribute(u'Node icon.')


class INavigationLeaf(ILeaf):
    """Marker interface for nodes which have no navigatable children."""


class IWorkflowState(INode):
    """Workflow support on nodes."""
    workflow_name = Attribute(u'Name of registered workflow.')
    workflow_tsf = Attribute(u'Translation string factory used to translate '
                             u'states and transitions')
    state = Attribute(u'Current workflow state.')


class IOwnerSupport(ISecured):
    """Interface for providing ownership information.

    Plumbs __acl__ property.
    """
    owner_attribute_name = Attribute(u'Attribute name of the owner field.')
    owner = Attribute(u'User id of node owner')


class IPrincipalACL(ISecured):
    """Principal specific roles on nodes.

    Plumbs __acl__ property.
    """
    role_inheritance = Attribute(u'Flag whether principal roles are '
                                 u'additionally aggregated from parent.')
    principal_roles = Attribute(u'Attribute containing principal roles for '
                                u'secured object.')
    aggregated_roles = Attribute(u'Aggregated roles.')

    def aggregated_roles_for(principal_id):
        """Return aggregated roles for principal by principal_id.
        """


class IACLAdapter(Interface):
    """Interface for providing ACL as adapter."""
    acl = Attribute(u'ACL')


class IAdapterACL(ISecured):
    """ACL from ``IACLAdapter`` on nodes."""
    default_acl = Attribute(u'Default ACL if no ``IAclAdapter`` found for node')


class ICopySupport(INode):
    """Copysupport for nodes."""

    supports_cut = Attribute(u'Supports cut')
    supports_copy = Attribute(u'Supports copy')
    supports_paste = Attribute(u'Supports paste')


class IUUIDAsName(IUUIDAware):
    """Exposes ``self.uuid`` as ``self.__name__``. Considers key changes in
    node trees at copy time.
    """


class ITranslation(ISchema):
    """A translation."""
    value = Attribute(u'The translated value according to the curren language')


class ILiveSearch(Interface):
    """Livesearch adapter."""

    def search(request, query):
        """Return search result for query.

        Return value is a list of dicts. Each result dict must contain at least
        the key ``value`` which contains the display value for the suggestions
        dropdown. Any other keys are optional, they are accessible in the JS
        callback when ``typeahead:selected`` gets triggered.
        """


class IAuthenticator(Interface):
    """Authenticator utility.

    If application ini file defines ``cone.authenticator`` setting, a named
    utility of IAuthenticator gets looked up and principal authentication
    happens against it.
    """

    def authenticate(login, password):
        """Authenticate principal.

        Return principal id if authentication is successful, else None.
        """



import sys
import unittest


def test_suite():
    from cone.app.tests import test_testing

    from cone.app.tests import test_app
    from cone.app.tests import test_model
    from cone.app.tests import test_security
    from cone.app.tests import test_ugm
    from cone.app.tests import test_utils
    from cone.app.tests import test_workflow

    from cone.app.tests import test_browser
    from cone.app.tests import test_browser_actions
    from cone.app.tests import test_browser_ajax
    from cone.app.tests import test_browser_authoring
    from cone.app.tests import test_browser_batch
    from cone.app.tests import test_browser_content
    from cone.app.tests import test_browser_contents
    from cone.app.tests import test_browser_contextmenu
    from cone.app.tests import test_browser_copysupport
    from cone.app.tests import test_browser_exception
    from cone.app.tests import test_browser_form
    from cone.app.tests import test_browser_layout
    from cone.app.tests import test_browser_login
    from cone.app.tests import test_browser_order
    from cone.app.tests import test_browser_referencebrowser
    from cone.app.tests import test_browser_resources
    from cone.app.tests import test_browser_settings
    from cone.app.tests import test_browser_sharing
    from cone.app.tests import test_browser_table
    from cone.app.tests import test_browser_translation
    from cone.app.tests import test_browser_utils
    from cone.app.tests import test_browser_workflow

    suite = unittest.TestSuite()

    suite.addTest(unittest.findTestCases(test_testing))

    suite.addTest(unittest.findTestCases(test_app))
    suite.addTest(unittest.findTestCases(test_model))
    suite.addTest(unittest.findTestCases(test_security))
    suite.addTest(unittest.findTestCases(test_ugm))
    suite.addTest(unittest.findTestCases(test_utils))
    suite.addTest(unittest.findTestCases(test_workflow))

    suite.addTest(unittest.findTestCases(test_browser))
    suite.addTest(unittest.findTestCases(test_browser_actions))
    suite.addTest(unittest.findTestCases(test_browser_ajax))
    suite.addTest(unittest.findTestCases(test_browser_authoring))
    suite.addTest(unittest.findTestCases(test_browser_batch))
    suite.addTest(unittest.findTestCases(test_browser_content))
    suite.addTest(unittest.findTestCases(test_browser_contents))
    suite.addTest(unittest.findTestCases(test_browser_contextmenu))
    suite.addTest(unittest.findTestCases(test_browser_copysupport))
    suite.addTest(unittest.findTestCases(test_browser_exception))
    suite.addTest(unittest.findTestCases(test_browser_form))
    suite.addTest(unittest.findTestCases(test_browser_layout))
    suite.addTest(unittest.findTestCases(test_browser_login))
    suite.addTest(unittest.findTestCases(test_browser_order))
    suite.addTest(unittest.findTestCases(test_browser_referencebrowser))
    suite.addTest(unittest.findTestCases(test_browser_resources))
    suite.addTest(unittest.findTestCases(test_browser_settings))
    suite.addTest(unittest.findTestCases(test_browser_sharing))
    suite.addTest(unittest.findTestCases(test_browser_table))
    suite.addTest(unittest.findTestCases(test_browser_translation))
    suite.addTest(unittest.findTestCases(test_browser_utils))
    suite.addTest(unittest.findTestCases(test_browser_workflow))

    return suite


def run_tests():
    from zope.testrunner.runner import Runner

    runner = Runner(found_suites=[test_suite()])
    runner.run()
    sys.exit(int(runner.failed))


if __name__ == '__main__':
    run_tests()



from cone.app import testing
from cone.app.interfaces import IWorkflowState
from cone.app.testing.mock import InexistentWorkflowNode
from cone.app.testing.mock import InterfaceWorkflowNode
from cone.app.testing.mock import StateACLWorkflowNode
from cone.app.testing.mock import WorkflowNode
from cone.app.workflow import initialize_workflow
from cone.app.workflow import lookup_state_data
from cone.app.workflow import lookup_workflow
from node.tests import NodeTestCase
from pyramid.security import ALL_PERMISSIONS
from repoze.workflow.workflow import Workflow


class TestWorkflow(NodeTestCase):
    layer = testing.security

    def test_lookup_workflow(self):
        wf = lookup_workflow(InexistentWorkflowNode)
        self.assertEqual(wf, None)

        wf = lookup_workflow(WorkflowNode)
        self.assertTrue(isinstance(wf, Workflow))

        wf = lookup_workflow(WorkflowNode())
        self.assertTrue(isinstance(wf, Workflow))

        wf = lookup_workflow(InterfaceWorkflowNode())
        self.assertTrue(isinstance(wf, Workflow))

        wf = lookup_workflow(InterfaceWorkflowNode)
        self.assertTrue(isinstance(wf, Workflow))

    def test_lookup_state_data(self):
        state_data = lookup_state_data(InexistentWorkflowNode)
        self.assertEqual(state_data, {})

        state_data = lookup_state_data(WorkflowNode())
        self.assertEqual(state_data, {
            'callback': None,
            'description': u'Foo',
            'title': u'Initial State'
        })

    def test_workflow(self):
        node = WorkflowNode()
        self.assertTrue(IWorkflowState.providedBy(node))

        # Workflow name is set on node properties for lookup
        self.assertEqual(node.workflow_name, u'dummy')

        # Initial workflow state gets set at node creation time if not set yet
        self.assertEqual(node.state, u'initial')

        node.state = u'final'
        self.assertEqual(node.attrs['state'], u'final')
        self.assertTrue(node.attrs['state'] is node.state)

        initialize_workflow(node)
        self.assertEqual(node.state, u'final')

    def test_copy(self):
        root = WorkflowNode()
        child = root['child'] = WorkflowNode()
        self.assertTrue(root.state == child.state == u'initial')

        root.state = child.state = u'final'
        self.assertTrue(root.state == child.state == u'final')

        # Workflow state gets set to initial state on copied nodes
        copied = root.copy()
        self.assertTrue(copied.state == copied['child'].state == u'initial')

    def test_acl(self):
        # Default workflow state ACL
        node = WorkflowNode()
        self.assertEqual(node.__acl__, [
            ('Allow', 'system.Authenticated', ['view']),
            ('Allow', 'role:viewer', ['view']),
            ('Allow', 'role:editor', ['view', 'add', 'edit']),
            ('Allow', 'role:owner', [
                'view', 'add', 'edit', 'delete',
                'change_state', 'manage_permissions'
            ]),
            ('Allow', 'role:admin', [
                'view', 'add', 'edit', 'delete',
                'change_state', 'manage_permissions'
            ]),
            ('Allow', 'role:manager', [
                'view', 'add', 'edit', 'delete', 'change_state',
                'manage_permissions', 'manage'
            ]),
            ('Allow', 'system.Everyone', ['login']),
            ('Deny', 'system.Everyone', ALL_PERMISSIONS)
        ])

        # If not set, and ACL not found in ``state_acls``, raise on access
        node.default_acl = None
        err = self.expect_error(ValueError, lambda: node.__acl__)
        self.assertEqual(str(err), "No ACL found for state 'initial'")

    def test_state_acl(self):
        node = StateACLWorkflowNode()
        self.assertEqual(node.workflow_name, u'dummy')
        self.assertTrue(IWorkflowState.providedBy(node))

        self.assertEqual(node.__acl__, [
            ('Allow', 'role:manager', ['manage', 'edit', 'change_state']),
            ('Allow', 'system.Everyone', ['login']),
            ('Deny', 'system.Everyone', ALL_PERMISSIONS)
        ])

        with self.layer.authenticated('manager'):
            request = self.layer.new_request()
            wf = lookup_workflow(node)
            wf.transition(node, request, u'initial_2_final')
            self.assertEqual(node.state, u'final')

        self.assertEqual(node.__acl__, [
            ('Allow', 'role:manager', ['view', 'edit', 'change_state']),
            ('Deny', 'system.Everyone', ALL_PERMISSIONS)
        ])



from cone.app import compat
from cone.app.utils import app_config
from cone.app.utils import format_traceback as _format_traceback
from cone.app.utils import node_path
from node.utils import safe_encode
from pyramid.i18n import TranslationStringFactory
import copy
import datetime
import re


_ = TranslationStringFactory('cone.app')


# B/C. use ``authenticated_userid`` directly.
def authenticated(request):
    return request.authenticated_userid


# B/C, removed as of cone.app 1.1
nodepath = node_path


# default query parameters to quote
QUOTE_PARAMS = ('came_from',)


def make_query(quote_params=QUOTE_PARAMS, **kw):
    query = list()
    for name, param in sorted(kw.items()):
        if param is None:
            continue
        if isinstance(param, compat.STR_TYPE):
            param = [param]
        if type(param) in compat.NUMBER_TYPES:
            param = [str(param)]
        quote = name in quote_params
        for p in param:
            p = safe_encode(p) if compat.IS_PY2 else p
            query.append('{}={}'.format(name, compat.quote(p) if quote else p))
    query = '&'.join(query)
    if query:
        return '?{}'.format(query)


def make_url(request, path=None, node=None, resource=None, query=None):
    # if path=[] in signature, path gets aggregated in recursive calls ???
    # happens on icon lookup in navtree.
    # ^^^ that is because the [] (a list, mutable) is generated at compile
    # time. mutable values should not be in function signatures to avoid this.
    if path is None:
        path = []
    else:
        path = copy.copy(path)
    if node is not None:
        path = node_path(node)
    if resource is not None:
        path.append(resource)
    path = [compat.quote(safe_encode(it)) for it in path]
    url = '{}/{}'.format(request.application_url, '/'.join(path))
    if not query:
        return url
    return '{}{}'.format(url, query)


def choose_name(container, name):
    name = re.sub(
        r'-{2,}', '-',
        re.sub(r'^\w-|-\w-|-\w$', '-',
               re.sub(r'\W', '-', name.strip()))).strip('-').lower()
    n = name
    i = 0
    while n in container:
        i += 1
        n = u'{}-{}'.format(name, i)
    return n.replace('/', '-').lstrip('+@')


def format_date(dt, long=True):
    if not isinstance(dt, datetime.datetime):
        return _('unknown', default='Unknown')
    return long and dt.strftime('%d.%m.%Y %H:%M') or dt.strftime('%d.%m.%Y')


def node_icon(node):
    if node.properties.icon:
        return node.properties.icon
    info = node.nodeinfo
    if not info.icon:
        return app_config().default_node_icon
    return info.icon


def request_property(func):
    """Decorator like ``property``, but underlying function is only called once
    per request.

    Cache attribute on request.environ under key
    ``instanceid.classname.funcname``.

    Works only on instances providing a request attribute.
    """
    def wrapper(self):
        cache_key = '{}.{}.{}'.format(
            str(id(self)),
            self.__class__.__name__,
            func.__name__
        )
        try:
            return self.request.environ[cache_key]
        except KeyError:
            val = self.request.environ[cache_key] = func(self)
            return val
    wrapper.__doc__ = func.__doc__
    return property(wrapper)


def format_traceback():
    return '<pre>{}</pre>'.format(_format_traceback())



from cone.app import testing
from cone.app.browser.resources import cone_css
from cone.app.browser.resources import cone_js
from cone.app.browser.resources import is_remote_resource
from cone.app.browser.resources import MergedAssets
from cone.app.browser.resources import print_css
from cone.app.browser.resources import Resources
from cone.app.model import Properties
from cone.app.testing.mock import static_resources
from cone.tile import render_tile
from cone.tile.tests import TileTestCase
import cone.app
import os
import pkg_resources


class TestBrowserResources(TileTestCase):
    layer = testing.security

    def test_is_remote_resource(self):
        self.assertTrue(is_remote_resource('http://example.com/foo'))
        self.assertTrue(is_remote_resource('https://example.com/foo'))
        self.assertTrue(is_remote_resource('//foo'))
        self.assertFalse(is_remote_resource('foo'))

    def test_MergedAssets(self):
        request = self.layer.new_request()
        assets = MergedAssets(request)

        assets.merged_js_assets = Properties()
        assets.merged_js_assets.public = [(static_resources, 'script1.js')]
        assets.merged_js_assets.protected = [(static_resources, 'script2.js')]

        assets.merged_css_assets = Properties()
        assets.merged_css_assets.public = [(static_resources, 'style1.css')]
        assets.merged_css_assets.protected = [(static_resources, 'style2.css')]

        assets.merged_print_css_assets = Properties()
        assets.merged_print_css_assets.public = [(static_resources, 'print1.css')]
        assets.merged_print_css_assets.protected = [(static_resources, 'print2.css')]

        expected = 'console.log("script1");\n\n'
        self.assertEqual(assets.merged_js, expected)
        expected = '.style1 { display: block; }\n\n'
        self.assertEqual(assets.merged_css, expected)
        expected = '.print1 { display: none; }\n\n'
        self.assertEqual(assets.merged_print_css, expected)

        with self.layer.authenticated('max'):
            res = assets.merged_js
        expected = 'console.log("script1");\n\nconsole.log("script2");\n\n'
        self.assertEqual(res, expected)

        with self.layer.authenticated('max'):
            res = assets.merged_css
        expected = '.style1 { display: block; }\n\n.style2 { display: block; }\n\n'
        self.assertEqual(res, expected)

        with self.layer.authenticated('max'):
            res = assets.merged_print_css
        expected = '.print1 { display: none; }\n\n.print2 { display: none; }\n\n'
        self.assertEqual(res, expected)

    def test_cone_js(self):
        model = cone.app.root
        request = self.layer.new_request()
        res = cone_js(model, request)
        self.assertEqual(res.headers['Content-Type'], 'application/javascript')

    def test_cone_css(self):
        model = cone.app.root
        request = self.layer.new_request()
        res = cone_css(model, request)
        self.assertEqual(res.headers['Content-Type'], 'text/css')

    def test_print_css(self):
        model = cone.app.root
        request = self.layer.new_request()
        res = print_css(model, request)
        self.assertEqual(res.headers['Content-Type'], 'text/css')

    def test_Resources(self):
        class DummyResources(Resources):
            @property
            def js(self):
                js = Properties()
                js.public = ['public.js', 'https://remote.foo/public.js']
                js.protected = ['protected.js', 'https://remote.foo/protected.js']
                return js

            @property
            def css(self):
                css = Properties()
                css.public = ['public.css', 'https://remote.foo/public.css']
                css.protected = ['protected.css', 'https://remote.foo/protected.css']
                return css

        resources = DummyResources()
        resources.model = cone.app.root
        resources.request = self.layer.new_request()

        self.assertEqual(resources.authenticated, None)
        with self.layer.authenticated('max'):
            self.assertEqual(resources.authenticated, 'max')

        self.assertEqual(
            resources.resource_url('foo.js'),
            'http://example.com/foo.js'
        )
        self.assertEqual(
            resources.resource_url('https://remote.foo/resource.js'),
            'https://remote.foo/resource.js'
        )

        res = resources.resources(resources.js)
        self.assertEqual(res, [
            'http://example.com/public.js',
            'https://remote.foo/public.js'
        ])

        with self.layer.authenticated('max'):
            res = resources.resources(resources.js)
        self.assertEqual(res, [
            'http://example.com/public.js',
            'https://remote.foo/public.js',
            'http://example.com/protected.js',
            'https://remote.foo/protected.js'
        ])

        res = resources.resources(resources.css)
        self.assertEqual(res, [
            'http://example.com/public.css',
            'https://remote.foo/public.css'
        ])

        with self.layer.authenticated('max'):
            res = resources.resources(resources.css)
        self.assertEqual(res, [
            'http://example.com/public.css',
            'https://remote.foo/public.css',
            'http://example.com/protected.css',
            'https://remote.foo/protected.css'
        ])

    def test_resources(self):
        # CSS Resource
        self.assertTrue(isinstance(cone.app.cfg.css, Properties))
        self.assertEqual(
            sorted(cone.app.cfg.css.keys()),
            ['protected', 'public']
        )

        # Contain CSS resources for authenticated users
        self.checkOutput('[...]', str(cone.app.cfg.css.protected))

        # Contain CSS resource for all users
        self.checkOutput('[...]', str(cone.app.cfg.css.public))

        # JS Resources
        self.assertTrue(isinstance(cone.app.cfg.js, Properties))
        self.assertEqual(
            sorted(cone.app.cfg.js.keys()),
            ['protected', 'public']
        )

        # Contain CSS resources for authenticated users
        self.checkOutput('[...]', str(cone.app.cfg.js.protected))

        # Contain CSS resource for all users
        self.checkOutput('[...]', str(cone.app.cfg.js.public))

        # Render resources tile unauthorized
        request = self.layer.new_request()
        result = render_tile(cone.app.root, request, 'resources')
        self.checkOutput('...<!-- javascripts -->...', result)

        # Render resources tile authorized
        with self.layer.authenticated('max'):
            result = render_tile(cone.app.root, request, 'resources')
        self.checkOutput('...<!-- javascripts -->...', result)

        # Merged Assets
        assets = cone.app.cfg.merged.js.public
        self.checkOutput("""
        [(<pyramid.static.static_view object at ...>, 'jquery-1.9.1.js'),
        (<pyramid.static.static_view object at ...>, 'jquery.migrate-1.2.1.js'),
        (<pyramid.static.static_view object at ...>, 'jqueryui/jquery-ui-1.10.3.custom.js'),
        (<pyramid.static.static_view object at ...>, 'bootstrap/js/bootstrap.js')...]
        """, str(assets))

        static = assets[0][0]
        resource = assets[0][1]
        self.assertEqual(static.package_name, 'cone.app.browser')
        self.assertEqual(static.docroot, 'static')

        subpath = os.path.join(static.docroot, resource)
        path = pkg_resources.resource_filename(static.package_name, subpath)
        self.checkOutput('/.../cone/app/browser/static/jquery-1.9.1.js', path)

        data = ''
        with open(path, 'r') as file:
            data += file.read() + '\n\n'
        self.checkOutput('...\n\n', data)

        request = self.layer.new_request()
        assets = MergedAssets(request)
        self.assertTrue(len(assets.merged_js) > 0)
        self.checkOutput('...', assets.merged_css)

        with self.layer.authenticated('admin'):
            self.assertTrue(len(assets.merged_js) > 0)
            self.checkOutput('...', assets.merged_css)



from setuptools import find_packages
from setuptools import setup
from setuptools.command.test import test
import os


def read_file(name):
    with open(os.path.join(os.path.dirname(__file__), name)) as f:
        return f.read()


version = '1.0.6'
shortdesc = 'Web application stub'
longdesc = '\n\n'.join([read_file(name) for name in [
    'README.rst',
    'CHANGES.rst',
    'LICENSE.rst'
]])


class Test(test):

    def run_tests(self):
        from cone.app import tests
        tests.run_tests()


setup(
    name='cone.app',
    version=version,
    description=shortdesc,
    long_description=longdesc,
    classifiers=[
        'Environment :: Web Environment',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    keywords='node pyramid cone web',
    author='Cone Contributors',
    author_email='dev@conestack.org',
    url='http://github.com/conestack/cone.app',
    license='Simplified BSD',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['cone'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'setuptools',
        'Chameleon',
        'node>=0.9.28',
        'node.ext.ugm>=0.9.13',
        'pyramid>=1.5',
        'pyramid_zcml',
        'pyramid_chameleon',
        'bdajax>=1.11',
        'cone.tile>=1.0',
        'yafowil>=2.3.4',
        'yafowil.webob',
        'yafowil.bootstrap',
        'repoze.workflow',
    ],
    extras_require=dict(
        lxml=[
            'lxml'
        ],
        yaml=[
            'yafowil.yaml'
        ],
        test=[
            'lxml',
            'yafowil.yaml',
            'zope.testrunner'
        ],
        docs=[
            'Sphinx',
            'sphinx_bootstrap_theme',
            'repoze.sphinx.autointerface'
        ],
    ),
    tests_require=[
        'lxml',
        'yafowil.yaml',
        'zope.testrunner'
    ],
    cmdclass=dict(test=Test),
    entry_points="""\
    [paste.app_factory]
    main = cone.app:main
    [paste.filter_app_factory]
    remote_addr = cone.app:make_remote_addr_middleware
    """
)



from cone.app import model
from cone.app.security import authenticate
from contextlib import contextmanager
from pyramid.security import AuthenticationAPIMixin
from pyramid.testing import DummyRequest as BaseDummyRequest
from pyramid.tests.test_view import DummyVenusianContext
from pyramid.tests.test_view import DummyVenusianInfo
from webob.acceptparse import create_accept_header
from zope.component import getGlobalSiteManager
from zope.component.hooks import resetHooks
from zope.configuration.xmlconfig import XMLConfig
import cone.app
import cone.tile
import os
import venusian


def reset_node_info_registry(fn):
    """Decorator for tests using node info registry
    """
    node_info_registry = model._node_info_registry
    def wrapper(*a, **kw):
        try:
            model._node_info_registry = dict()
            fn(*a, **kw)
        finally:
            model._node_info_registry = node_info_registry
    return wrapper


class DummyRequest(BaseDummyRequest, AuthenticationAPIMixin):
    _accept = None

    @property
    def is_xhr(self):
        return self.headers.get('X-Requested-With') == 'XMLHttpRequest'

    @property
    def accept(self):
        if self._accept is None:
            self._accept = create_accept_header(None)
        return self._accept

    @accept.setter
    def accept(self, value):
        self._accept = create_accept_header(value)

    @accept.deleter
    def accept(self):
        self._accept = None


class DummyVenusian(object):

    def __init__(self):
        self.attachments = []

    def attach(self, wrapped, callback, category=None, depth=1):
        self.attachments.append((wrapped, callback))
        return DummyVenusianInfo()

    def call_attachemnts(self):
        for wrapped, callback in self.attachments:
            callback(DummyVenusianContext(), None, wrapped)


DATADIR = os.path.join(os.path.dirname(__file__), 'data', 'ugm')


class Security(object):
    """Test layer with dummy authentication for security testing.
    """
    current_request = None
    auth_env_keys = [
        'REMOTE_USER_TOKENS',
        'REMOTE_USER_DATA',
        'HTTP_COOKIE',
        'cone.app.user.roles',
    ]

    def __init__(self):
        self.__name__ = self.__class__.__name__
        self.__bases__ = []

    @property
    def registry(self):
        return getGlobalSiteManager()

    def defaults(self):
        return {
            'request': self.current_request,
            'registry': self.registry
        }

    def new_request(self, type=None, xhr=False):
        request = self.current_request
        auth = dict()
        cookies = dict()
        if request:
            cookies = request.cookies
            environ = request.environ
            for key in self.auth_env_keys:
                if key in environ:
                    auth[key] = environ[key]
        request = DummyRequest()
        request.cookies.update(cookies)
        request.environ['SERVER_NAME'] = 'testcase'
        request.environ['AUTH_TYPE'] = 'cookie'
        request.environ.update(auth)
        request.params['_LOCALE_'] = 'en'
        request.accept = 'text/html'
        if type == 'json':
            request.headers['X-Request'] = 'JSON'
            request.accept = 'application/json'
        if xhr:
            request.headers['X-Requested-With'] = 'XMLHttpRequest'
        self.current_request = request
        return request

    def forget_request(self):
        self.current_request = None

    def set_lang(self, lang):
        request = self.current_request
        if not request:
            request = self.new_request()
        request.params['_LOCALE_'] = lang

    def login(self, login, password=None):
        request = self.current_request
        if not request:
            request = self.new_request()
        else:
            self.logout()
        password = password if password else 'secret'
        res = authenticate(request, login, password)
        if res:
            request.environ['HTTP_COOKIE'] = res[0][1]
            cookie = res[0][1].split(';')[0].split('=')
            request.cookies[cookie[0]] = cookie[1]

    def logout(self):
        request = self.current_request
        if request:
            request.cookies.clear()
            environ = request.environ
            for key in self.auth_env_keys:
                if key in environ:
                    del environ[key]

    @contextmanager
    def authenticated(self, login, password=None):
        try:
            self.login(login, password=password)
            yield
        finally:
            self.logout()

    @contextmanager
    def hook_tile_reg(self):
        venusian_ = DummyVenusian()
        try:
            cone.tile.tile.venusian = venusian_
            yield
        finally:
            venusian_.call_attachemnts()
            cone.tile.tile.venusian = venusian

    def make_app(self, **kw):
        import pyramid.threadlocal
        pyramid.threadlocal.manager.default = self.defaults
        settings = {
            'default_locale_name': 'en',
            'cone.available_languages': 'en, de',
            'cone.admin_user': 'superuser',
            'cone.admin_password': 'superuser',
            'cone.auth_secret': '12345',
            'cone.plugins': 'node.ext.ugm\ninexistent',
            'cone.root.title': 'cone',
            'cone.root.default_child': None,
            'cone.root.default_content_tile': 'content',
            'cone.root.mainmenu_empty_title': False,
            'ugm.backend': 'file',
            'ugm.user_display_attr': 'fullname',
            'ugm.group_display_attr': 'groupname',
            'ugm.users_file': os.path.join(DATADIR, 'users'),
            'ugm.groups_file': os.path.join(DATADIR, 'groups'),
            'ugm.roles_file': os.path.join(DATADIR, 'roles'),
            'ugm.datadir': os.path.join(DATADIR, 'userdata'),
            'ugm.user_expires_attr': 'expires',
            'testing.hook_global_registry': True,
        }
        settings.update(**kw)
        self.app = cone.app.main({}, **settings)
        self.current_request = None

    def setUp(self, args=None):
        self.make_app()
        XMLConfig('testing/dummy_workflow.zcml', cone.app)()
        print("Security set up.")

    def tearDown(self):
        # XXX: something is wrong here.
        import pyramid.threadlocal
        pyramid.threadlocal.manager.default = pyramid.threadlocal.defaults
        resetHooks()
        print("Security torn down.")


security = Security()



from cone.app.browser.utils import make_query
from cone.app.browser.utils import make_url
from cone.app.interfaces import ICopySupport
from cone.app.interfaces import IPrincipalACL
from cone.app.interfaces import IWorkflowState
from cone.tile import render_template
from cone.tile import render_tile
from node.interfaces import IOrder
from odict import odict
from pyramid.i18n import TranslationStringFactory


_ = TranslationStringFactory('cone.app')


def get_action_context(request):
    return request.environ['action_context']


class ActionContext(object):
    """The action context is used to determine action scopes. The action scope
    is used by browser actions to check it's own state, e.g. if action button
    is selected, disabled or displayed at all.
    """

    def __init__(self, model, request, tilename):
        self.model = model
        self.request = request
        self.tilename = tilename
        request.environ['action_context'] = self

    @property
    def scope(self):
        model = self.model
        request = self.request
        scope = self.tilename
        # if ``bdajax.action`` found on request, use it as current scope
        if 'bdajax.action' in request.params:
            scope = request.params['bdajax.action']
        # if action is ``layout``, content tile name is passed
        if scope == 'layout':
            scope = request.params.get('contenttile', 'content')
        # change model if default child defined
        if model.properties.default_child:
            model = model[model.properties.default_child]
        # change scope if default content rendering and custom default
        # content tile
        if model.properties.default_content_tile and scope == 'content':
            scope = model.properties.default_content_tile
        return scope


class Toolbar(odict):
    """A toolbar rendering actions.
    """
    display = True
    css = None

    def __init__(self, **kw):
        self.__dict__.update(kw)

    def __call__(self, model, request):
        if not self.display:
            return u''
        rendered_actions = list()
        for action in self.values():
            rendered = action(model, request)
            if not rendered:
                continue
            rendered_actions.append(rendered)
        if not rendered_actions:
            return u''
        rendered_actions = u'\n'.join(rendered_actions)
        if not self.css:
            return u'<div>%s</div>' % rendered_actions
        return u'<div class="%s">%s</div>' % (self.css, rendered_actions)


class Action(object):
    """Abstract Action.
    """
    display = True

    def __call__(self, model, request):
        self.model = model
        self.request = request
        if not self.display:
            return u''
        return self.render()

    @property
    def action_scope(self):
        action_context = self.request.environ.get('action_context', None)
        if action_context is None:
            return ''
        return action_context.scope

    def permitted(self, permission):
        return self.request.has_permission(permission, self.model)

    def render(self):
        raise NotImplementedError(u"Abstract ``Action`` does not implement "
                                  u"render.")


class TileAction(Action):
    """Action rendered by a tile.
    """
    tile = u''

    def render(self):
        return render_tile(self.model, self.request, self.tile)


class TemplateAction(Action):
    """Action rendered by a template.
    """
    template = u''

    def render(self):
        return render_template(
            self.template,
            request=self.request,
            model=self.model,
            context=self
        )


class DropdownAction(TemplateAction):
    """Action rendering a dropdown.
    """
    template = u'cone.app.browser:templates/action_dropdown.pt'
    href = None
    css = None
    title = None

    @property
    def items(self):
        """Return list of ``cone.app.model.Properties`` instances providing
        attributes ``icon``, ``url``, ``target``, ``action`` and ``title``.
        """
        raise NotImplementedError(u"Abstract ``DropdownAction`` does not "
                                  u"implement  ``items``")


class LinkAction(TemplateAction):
    """Action rendering a HTML link, optional with bdajax attributes.
    """
    template = 'cone.app.browser:templates/link_action.pt'
    bind = 'click'       # ajax:bind attribute
    id = None            # id attribute
    href = '#'           # href attribute
    css = None           # in addition for computed class attribute
    title = None         # title attribute
    action = None        # ajax:action attribute
    event = None         # ajax:event attribute
    confirm = None       # ajax:confirm attribute
    overlay = None       # ajax:overlay attribute
    path = None          # ajax:path attribute
    path_target = None   # ajax:path-target attribute
    path_action = None   # ajax:path-action attribute
    path_event = None    # ajax:path-event attribute
    path_overlay = None  # ajax:path-overlay attribute
    text = None          # link text
    enabled = True       # if false, link gets 'disabled' css class
    selected = False     # if true, link get 'selected' css class
    icon = None          # if set, add span tag with value as CSS class in link

    def __init__(self, **kw):
        self.__dict__.update(kw)

    @property
    def css_class(self):
        css = not self.enabled and 'disabled' or ''
        css = self.selected and '%s selected' % css or css
        if self.css:
            css = '%s %s' % (self.css, css)
        css = css.strip()
        return css and css or None

    @property
    def target(self):
        return make_url(self.request, node=self.model)


class ButtonAction(TemplateAction):
    """Action rendering a HTML button, optional with bdajax attributes.
    """
    template = 'cone.app.browser:templates/button_action.pt'
    bind = 'click'         # ajax:bind attribute
    id = None              # id attribute
    css = None             # in addition for computed class attribute
    title = None           # title attribute
    type = None            # type of button
    name = None            # name for the button
    value = None           # initial value for the button
    autofocus = None       # button gets focus on page load
    disabled = None        # button should be disabled
    form = None            # form the button belongs to
    formaction = None      # where to send the data when form is submitted
    formenctype = None     # how form-data should be encoded before sending
    formmethod = None      # specifies http method
    formnovalidate = None  # data should not be validated on submission
    formtarget = None      # where to display response after form submission
    action = None          # ajax:action attribute
    event = None           # ajax:event attribute
    confirm = None         # ajax:confirm attribute
    overlay = None         # ajax:overlay attribute
    path = None            # ajax:path attribute
    path_target = None     # ajax:path-target attribute
    path_action = None     # ajax:path-action attribute
    path_event = None      # ajax:path-event attribute
    path_overlay = None    # ajax:path-overlay attribute
    text = None            # button text
    icon = None            # if set, add i tag with value as CSS class

    def __init__(self, **kw):
        self.__dict__.update(kw)

    @property
    def target(self):
        return make_url(self.request, node=self.model)


class ActionUp(LinkAction):
    """One level up action.
    """
    id = 'toolbaraction-up'
    icon = 'glyphicon glyphicon-arrow-up'
    event = 'contextchanged:#layout'
    text = _('action_one_level_up', default='One level up')
    path = 'href'

    @property
    def display(self):
        return self.model.properties.action_up \
            and self.request.has_permission('view', self.model.parent) \
            and self.permitted('view')

    @property
    def href(self):
        return make_url(self.request, node=self.container)

    @property
    def target(self):
        contenttile = self.model.properties.action_up_tile
        if not contenttile:
            contenttile = 'listing'
        query = make_query(contenttile=contenttile)
        return make_url(self.request, node=self.container, query=query)

    @property
    def container(self):
        container = self.model.parent
        default_child = container.properties.default_child
        if default_child and self.model.name == default_child:
            container = container.parent
        return container


class ActionView(LinkAction):
    """View action.
    """
    id = 'toolbaraction-view'
    icon = 'glyphicon glyphicon-eye-open'
    text = _('action_view', default='View')
    href = LinkAction.target
    path = 'href'

    @property
    def action(self):
        # XXX: use layout:#layout:replace
        # XXX: consider related view if set?
        contenttile = 'content'
        if self.model.properties.default_content_tile:
            contenttile = 'view'
        return '%s:#content:inner' % contenttile

    @property
    def display(self):
        return self.model.properties.action_view and self.permitted('view')

    @property
    def selected(self):
        if self.model.properties.default_content_tile:
            return self.action_scope == 'view'
        return self.action_scope == 'content'


class ViewLink(ActionView):
    """View link
    """
    css = None
    icon = None

    @property
    def text(self):
        return self.model.metadata.get('title', self.model.name)

    @property
    def display(self):
        return self.permitted('view')


class ActionList(LinkAction):
    """Contents listing action.
    """
    id = 'toolbaraction-list'
    icon = 'glyphicon glyphicon-th-list'
    action = 'listing:#content:inner'
    path = 'href'
    text = _('action_listing', default='Listing')

    @property
    def href(self):
        return '{}/listing'.format(make_url(self.request, node=self.model))

    @property
    def display(self):
        return self.model.properties.action_list and self.permitted('view')

    @property
    def selected(self):
        return self.action_scope == 'listing'


class ActionSharing(LinkAction):
    """Sharing action.
    """
    id = 'toolbaraction-share'
    icon = 'glyphicon glyphicon-share'
    action = 'sharing:#content:inner'
    path = 'href'
    text = _('action_sharing', default='Sharing')

    @property
    def href(self):
        return '{}/sharing'.format(make_url(self.request, node=self.model))

    @property
    def display(self):
        return IPrincipalACL.providedBy(self.model) \
            and self.permitted('manage_permissions')

    @property
    def selected(self):
        return self.action_scope == 'sharing'


class ActionState(TileAction):
    """Change state action.
    """
    tile = 'wf_dropdown'

    @property
    def display(self):
        return IWorkflowState.providedBy(self.model) \
            and self.permitted('change_state')


class ActionAdd(TileAction):
    """Add dropdown action.
    """
    tile = 'add_dropdown'

    @property
    def display(self):
        return self.permitted('add') and self.model.nodeinfo.addables


class ActionEdit(LinkAction):
    """Edit action.
    """
    id = 'toolbaraction-edit'
    icon = 'glyphicon glyphicon-pencil'
    action = 'edit:#content:inner'
    path = 'href'
    text = _('action_edit', default='Edit')

    @property
    def href(self):
        return '{}/edit'.format(make_url(self.request, node=self.model))

    @property
    def display(self):
        return self.model.properties.action_edit and self.permitted('edit')

    @property
    def selected(self):
        return self.action_scope == 'edit'


class ActionDelete(LinkAction):
    """Delete action.
    """
    id = 'toolbaraction-delete'
    icon = 'ion-trash-a'
    action = 'delete:NONE:NONE'
    confirm = _('delete_item_confirm',
                default='Do you really want to delete this Item?')
    text = _('action_delete', default='Delete')

    @property
    def display(self):
        # XXX: scope in subclass for contextmenu
        scope = self.action_scope == 'content'
        if self.model.properties.default_content_tile:
            scope = self.action_scope == 'view'
        return self.model.properties.action_delete \
            and self.permitted('delete') \
            and scope


class ActionDeleteChildren(LinkAction):
    """Delete children action.
    """
    id = 'toolbaraction-delete-children'
    icon = 'ion-trash-a'
    action = 'delete_children:NONE:NONE'
    confirm = _('delete_items_confirm',
                default='Do you really want to delete selected Items?')
    text = _('action_delete_selected_children',
             default='Delete selected children')

    @property
    def display(self):
        return self.model.properties.action_delete_children \
            and self.permitted('delete')

    @property
    def enabled(self):
        return self.request.cookies.get('cone.app.selected')


class ActionCut(LinkAction):
    """Cut children action.
    """
    id = 'toolbaraction-cut'
    icon = 'ion-scissors'
    text = _('action_cut', default='Cut')
    bind = None

    @property
    def display(self):
        return ICopySupport.providedBy(self.model) \
            and self.model.supports_cut \
            and self.permitted('cut') \
            and self.action_scope == 'listing'


class ActionCopy(LinkAction):
    """Copy children action.
    """
    id = 'toolbaraction-copy'
    icon = 'ion-ios7-copy-outline'
    text = _('action_copy', default='Copy')
    bind = None

    @property
    def display(self):
        return ICopySupport.providedBy(self.model) \
            and self.model.supports_copy \
            and self.permitted('copy') \
            and self.action_scope == 'listing'


class ActionPaste(LinkAction):
    """Paste children action.
    """
    id = 'toolbaraction-paste'
    icon = 'ion-clipboard'
    text = _('action_paste', default='Paste')
    bind = None

    @property
    def display(self):
        return ICopySupport.providedBy(self.model) \
            and self.model.supports_paste \
            and self.permitted('paste') \
            and self.action_scope == 'listing'

    @property
    def enabled(self):
        return self.request.cookies.get('cone.app.copysupport.cut') \
            or self.request.cookies.get('cone.app.copysupport.copy')


class _ActionMove(LinkAction):
    """Basic move action.
    """

    @property
    def display(self):
        if self.request.params.get('sort'):
            return False
        parent = self.model.parent
        return parent.properties.action_move \
            and IOrder.providedBy(parent) \
            and self.request.has_permission('change_order', parent)

    @property
    def target(self):
        request = self.request
        query = make_query(
            b_page=request.params.get('b_page'),
            size=request.params.get('size')
        )
        return make_url(self.request, node=self.model, query=query)


class ActionMoveUp(_ActionMove):
    """Move up action.
    """
    id = 'toolbaraction-move-up'
    icon = 'glyphicon glyphicon-chevron-up'
    action = 'move_up:NONE:NONE'
    text = _('move_up', default='Move up')

    @property
    def display(self):
        if not super(ActionMoveUp, self).display:
            return False
        return self.model.parent.first_key != self.model.name


class ActionMoveDown(_ActionMove):
    """Move down action.
    """
    id = 'toolbaraction-move-down'
    icon = 'glyphicon glyphicon-chevron-down'
    action = 'move_down:NONE:NONE'
    text = _('move_down', default='Move down')

    @property
    def display(self):
        if not super(ActionMoveDown, self).display:
            return False
        return self.model.parent.last_key != self.model.name



from cone.app.browser.actions import ActionAdd
from cone.app.browser.actions import ActionCopy
from cone.app.browser.actions import ActionCut
from cone.app.browser.actions import ActionDelete
# from cone.app.browser.actions import ActionDeleteChildren
from cone.app.browser.actions import ActionEdit
from cone.app.browser.actions import ActionList
from cone.app.browser.actions import ActionPaste
from cone.app.browser.actions import ActionSharing
from cone.app.browser.actions import ActionState
from cone.app.browser.actions import ActionUp
from cone.app.browser.actions import ActionView
from cone.app.browser.actions import LinkAction
from cone.app.browser.actions import TemplateAction
from cone.app.browser.actions import Toolbar
from cone.tile import Tile
from cone.tile import render_template
from cone.tile import tile
from odict import odict
from pyramid.i18n import TranslationStringFactory


_ = TranslationStringFactory('cone.app')


class ContextMenuToolbar(Toolbar):

    def __call__(self, model, request):
        if not self.display:
            return u''
        rendered_actions = list()
        for action in self.values():
            rendered = action(model, request)
            if not rendered:
                continue
            # expect correct markup if no link action
            if not isinstance(action, LinkAction):
                rendered_actions.append(rendered)
                continue
            # wrap link action in list item
            if action.selected:
                rendered = u'<li class="active">%s</li>' % rendered
            else:
                rendered = u'<li>%s</li>' % rendered
            rendered_actions.append(rendered)
        if not rendered_actions:
            return u''
        rendered_actions = u'\n'.join(rendered_actions)
        css = u'nav navbar-nav'
        if self.css:
            css += ' ' + self.css
        return u'<ul class="%s">%s</ul>' % (css, rendered_actions)


class ContextMenuDropdown(Toolbar):
    template = 'cone.app.browser:templates/contextmenu_dropdown.pt'
    icon = None
    title = None

    @property
    def display(self):
        for val in self.values():
            val.model = self.model
            val.request = self.request
            if val.display:
                return True
        return False

    def __call__(self, model, request):
        self.model = model
        self.request = request
        return render_template(
            self.template,
            request=request,
            model=model,
            context=self
        )

    def __setitem__(self, name, value):
        if not isinstance(value, TemplateAction):
            raise ValueError(u'Only ``TemplateAction`` deriving objects can be '
                             u'added to ``ContextMenuDropdown`` instances.')
        super(ContextMenuDropdown, self).__setitem__(name, value)


# context menu group registry
context_menu = odict()


class context_menu_group(object):
    """Decorator defining a context menu group.
    """

    def __init__(self, name):
        self.name = name

    def __call__(self, factory):
        context_menu[self.name] = factory()
        return factory


class context_menu_item(object):
    """Decorator defining a context menu item inside a group.
    """

    def __init__(self, group, name):
        self.group = group
        self.name = name

    def __call__(self, factory):
        context_menu[self.group][self.name] = factory()
        return factory


@context_menu_group(name='navigation')
class NavigationToolbar(ContextMenuToolbar):
    """Context menu navigation toolbar.
    """


context_menu_item(group='navigation', name='up')(ActionUp)


@context_menu_group(name='contentviews')
class ContentViewsDropdown(ContextMenuDropdown):
    """Context menu content views dropdown.
    """
    title = _('display', default=u'Display')


context_menu_item(group='contentviews', name='list')(ActionList)
context_menu_item(group='contentviews', name='view')(ActionView)
context_menu_item(group='contentviews', name='edit')(ActionEdit)
context_menu_item(group='contentviews', name='sharing')(ActionSharing)


@context_menu_group(name='childactions')
class ChildActionsDropdown(ContextMenuDropdown):
    """Context menu content views dropdown.
    """
    title = _('actions', default=u'Actions')


context_menu_item(group='childactions', name='cut')(ActionCut)
context_menu_item(group='childactions', name='copy')(ActionCopy)
context_menu_item(group='childactions', name='paste')(ActionPaste)
# context_menu_item(group='childactions', name='delete')(ActionDeleteChildren)


@context_menu_group(name='contextactions')
class ContextActionsToolbar(ContextMenuToolbar):
    """Context menu navigation toolbar.
    """


context_menu_item(group='contextactions', name='change_state')(ActionState)
context_menu_item(group='contextactions', name='add')(ActionAdd)
context_menu_item(group='contextactions', name='delete')(ActionDelete)


@tile(name='contextmenu', path='templates/contextmenu.pt', permission='view')
class ContextMenu(Tile):

    @property
    def toolbars(self):
        return context_menu.values()



from node.ext.ugm.file import Ugm as FileUgm
import logging


logger = logging.getLogger('cone.app')


class ugm_backend(object):
    """UGM backend configuration.
    """
    registry = dict()
    name = None
    factory = None
    ugm = None
    user_display_attr = 'fullname'
    group_display_attr = 'groupname'

    def __init__(self, name):
        self._reg_name = name

    def __call__(self, factory):
        self.registry[self._reg_name] = factory
        return factory

    @classmethod
    def load(cls, name, settings):
        if name not in cls.registry:
            raise ValueError('Unknown UGM backend "{}"'.format(name))
        cls.name = name
        cls.factory = cls.registry[name](settings)

    @classmethod
    def initialize(cls):
        if not cls.factory:
            raise ValueError('UGM backend not loaded')
        cls.ugm = cls.factory()


class UGMFactory(object):
    """UGM backend factory.
    """

    def __init__(self, settings):
        """Gets called by ``ugm_backend.load`` and is responsible to read
        UGM related configuration from passed ``settings`` dict.
        """
        raise NotImplementedError(
            'Abstract ``UGMFactory`` does not implement ``__init__``')

    def __call__(self):
        """Gets calles by ``ugm_backend.initialize`` and is responsible to
        instanciate and return a concrete ``node.ext.ugm.Ugm`` implementation.
        """
        raise NotImplementedError(
            'Abstract ``UGMFactory`` does not implement ``__call__``')


@ugm_backend('file')
class FileUGMFactory(UGMFactory):
    """UGM backend factory for file based UGM implementation.
    """

    def __init__(self, settings):
        self.users_file = settings.get('ugm.users_file')
        self.groups_file = settings.get('ugm.groups_file')
        self.roles_file = settings.get('ugm.roles_file')
        self.datadir = settings.get('ugm.datadir')
        self.user_expires_attr = settings.get('ugm.user_expires_attr')

    def __call__(self):
        return FileUgm(
            name='ugm',
            users_file=self.users_file,
            groups_file=self.groups_file,
            roles_file=self.roles_file,
            data_directory=self.datadir,
            user_expires_attr=self.user_expires_attr,
        )


@ugm_backend('node.ext.ugm')
class BCFileUGMFactory(FileUGMFactory):
    """B/C factory as replacement for the removed main hook from
    ``node.ext.ugm``. Actually the same as ``FileUGMFactory`` but reads
    settings from different names.
    """

    def __init__(self, settings):
        self.users_file = settings.get('node.ext.ugm.users_file')
        self.groups_file = settings.get('node.ext.ugm.groups_file')
        self.roles_file = settings.get('node.ext.ugm.roles_file')
        self.datadir = settings.get('node.ext.ugm.datadir')
        self.user_expires_attr = settings.get('node.ext.ugm.user_expires_attr')


def principal_data(principal_id):
    data = dict()
    ugm = ugm_backend.ugm
    try:
        user = ugm.users.get(principal_id)
        if not user:
            return data
        data = user.attrs
    except Exception as e:
        logger.error(str(e))
    return data



from cone.app import testing
from cone.app.browser.order import MoveAction
from cone.app.model import AppNode
from cone.app.model import BaseNode
from cone.tile import render_tile
from cone.tile.tests import TileTestCase
from node.behaviors import DefaultInit
from node.behaviors import DictStorage
from node.behaviors import MappingAdopt
from node.behaviors import MappingNode
from node.behaviors import Order
from node.interfaces import IOrder
from plumber import plumbing


class TestBrowserOrder(TileTestCase):
    layer = testing.security

    def test_moving(self):
        move_action = MoveAction()
        with self.assertRaises(NotImplementedError):
            move_action.move()

        @plumbing(
            AppNode,
            MappingAdopt,
            DefaultInit,
            MappingNode,
            DictStorage)
        class UnorderedNode(object):
            pass

        node = UnorderedNode()
        node['child'] = BaseNode()

        self.assertFalse(IOrder.providedBy(node))
        with self.layer.authenticated('manager'):
            request = self.layer.new_request()
            self.assertEqual(render_tile(node['child'], request, 'move_up'), u'')
        self.assertEqual(
            request.environ['cone.app.continuation'][0].payload,
            u'Object "child" not movable'
        )

        @plumbing(Order)
        class OrderableNode(BaseNode):
            def __call__(self):
                pass

        node = OrderableNode()
        node['a'] = BaseNode()
        node['b'] = BaseNode()

        self.assertEqual(node.properties.action_move, None)
        with self.layer.authenticated('manager'):
            request = self.layer.new_request()
            self.assertEqual(render_tile(node['b'], request, 'move_up'), u'')
        self.assertEqual(
            request.environ['cone.app.continuation'][0].payload,
            u'You are not permitted to move this object'
        )

        node.properties.action_move = True
        with self.layer.authenticated('manager'):
            request = self.layer.new_request()
            self.assertEqual(render_tile(node['a'], request, 'move_down'), u'')
        self.assertEqual(node.keys(), ['b', 'a'])

        with self.layer.authenticated('manager'):
            request = self.layer.new_request()
            self.assertEqual(render_tile(node['a'], request, 'move_up'), u'')
        self.assertEqual(node.keys(), ['a', 'b'])



# -*- coding: utf-8 -*-
from cone.app import browser
from cone.app import security
from cone.app.interfaces import IApplicationNode
from cone.app.model import AppRoot
from cone.app.model import AppSettings
from cone.app.model import LayoutConfig
from cone.app.model import Properties
from cone.app.ugm import ugm_backend
from cone.app.utils import format_traceback
from cone.app.utils import node_path
from node.interfaces import INode
from pyramid.authentication import AuthTktAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.config import Configurator
from pyramid.static import static_view
from pyramid.traversal import ResourceTreeTraverser
from yafowil.resources import YafowilResources as YafowilResourcesBase
from zope.component import adapter
from zope.component import getGlobalSiteManager
import importlib
import logging
import pyramid_chameleon
import pyramid_zcml
import sys
import threading


logger = logging.getLogger('cone.app')

# configuration
cfg = Properties()

# available languages
cfg.available_languages = []

# used main template
cfg.main_template = 'cone.app.browser:templates/main.pt'

# default node icon
cfg.default_node_icon = 'glyphicon glyphicon-asterisk'

# XXX: move resource registration to browser package
# XXX: support developmenet and production mode

# JS resources
cfg.js = Properties()
cfg.js.public = [
    'static/public.js'
]
cfg.js.protected = [
    'static/protected.js'
]

# CSS Resources
cfg.css = Properties()

# development
cfg.css.public = [
    'static/jqueryui/jquery-ui-1.10.3.custom.css',
    'static/bootstrap/css/bootstrap.css',
    'static/bootstrap/css/bootstrap-theme.css',
    'static/ionicons/css/ionicons.css',
    'static/typeahead/typeahead.css',
    '++resource++bdajax/bdajax_bs3.css',
    'static/styles.css'
]

# production
# cfg.css.public = [
#     'static/jqueryui/jquery-ui-1.10.3.custom.css',
#     'static/bootstrap/css/bootstrap.min.css',
#     'static/bootstrap/css/bootstrap-theme.min.css',
#     'static/ionicons/css/ionicons.css',
#     'static/typeahead/typeahead.css',
#     '++resource++bdajax/bdajax_bs3.css',
#     'static/styles.css'
# ]

cfg.css.protected = list()

# JS and CSS Assets to publish merged
cfg.merged = Properties()
cfg.merged.js = Properties()

# development
cfg.merged.js.public = [
    (browser.static_resources, 'jquery-1.9.1.js'),
    (browser.static_resources, 'jquery.migrate-1.2.1.js'),
    (browser.static_resources, 'jqueryui/jquery-ui-1.10.3.custom.js'),
    (browser.static_resources, 'bootstrap/js/bootstrap.js'),
    (browser.static_resources, 'typeahead/typeahead.bundle.js'),
    (browser.static_resources, 'cookie_functions.js')
]

# production
# cfg.merged.js.public = [
#     (browser.static_resources, 'jquery-1.9.1.min.js'),
#     (browser.static_resources, 'jquery.migrate-1.2.1.min.js'),
#     (browser.static_resources, 'jqueryui/jquery-ui-1.10.3.custom.min.js'),
#     (browser.static_resources, 'bootstrap/js/bootstrap.min.js'),
#     (browser.static_resources, 'typeahead/typeahead.bundle.js'),
#     (browser.static_resources, 'cookie_functions.js')
# ]

cfg.merged.js.protected = list()

cfg.merged.css = Properties()
cfg.merged.css.public = list()
cfg.merged.css.protected = list()

cfg.merged.print_css = Properties()
cfg.merged.print_css.public = [
    (browser.static_resources, 'print.css')
]
cfg.merged.print_css.protected = list()


class layout_config(object):
    _registry = dict()

    def __init__(self, *for_):
        self.for_ = for_

    def __call__(self, factory):
        for context in self.for_:
            self._registry[context] = factory
        return factory

    @classmethod
    def lookup(cls, model=None, request=None):
        for cls_ in model.__class__.mro():
            factory = cls._registry.get(cls_)
            if factory:
                return factory(model=model, request=request)


@layout_config(object)
class DefaultLayoutConfig(LayoutConfig):

    def __init__(self, model=None, request=None):
        super(DefaultLayoutConfig, self).__init__(model=model, request=request)
        self.mainmenu = True
        self.mainmenu_fluid = False
        self.livesearch = True
        self.personaltools = True
        self.columns_fluid = False
        self.pathbar = True
        self.sidebar_left = ['navtree']
        self.sidebar_left_grid_width = 3
        self.content_grid_width = 9


def import_from_string(path):
    mod, ob = path.rsplit('.', 1)
    return getattr(importlib.import_module(mod), ob)


root = None


def get_root(environ=None):
    return root


def configure_root(root, settings):
    root.metadata.title = settings.get('cone.root.title', 'CONE')
    root.properties.default_child = settings.get('cone.root.default_child')
    mainmenu_empty_title = settings.get('cone.root.mainmenu_empty_title')
    mainmenu_empty_title = mainmenu_empty_title in ['True', 'true', '1']
    root.properties.mainmenu_empty_title = mainmenu_empty_title
    default_content_tile = settings.get('cone.root.default_content_tile')
    if default_content_tile:
        root.properties.default_content_tile = default_content_tile
    root.properties.in_navtree = False


def default_root_node_factory(settings):
    root = AppRoot()
    root.factories['settings'] = AppSettings
    configure_root(root, settings)
    return root


def register_config(key, factory):
    root = get_root()
    factories = root['settings'].factories
    if key in factories:
        raise ValueError(u"Config with name '%s' already registered." % key)
    factories[key] = factory


# B/C
register_plugin_config = register_config


def register_entry(key, factory):
    root = get_root()
    factories = root.factories
    if key in factories:
        raise ValueError(u"Entry with name '%s' already registered." % key)
    root.factories[key] = factory


# B/C
register_plugin = register_entry


main_hooks = list()


def main_hook(func):
    """decorator to register main hook.

    Decorated function gets called on application startup.
    """
    main_hooks.append(func)
    return func


# B/C
def register_main_hook(callback):
    """Register function to get called on application startup.
    """
    main_hooks.append(callback)


thread_shutdown_hooks = list()


def thread_shutdown_hook(func):  # pragma: no cover
    """decorator to register thread shutdown hook.

    Decorated function gets called when main thread joins. Thread shutdown
    hooks are used for graceful joining of non daemon threads.
    """
    thread_shutdown_hooks.append(func)
    return func


def auth_tkt_factory(**kwargs):
    kwargs.setdefault('callback', security.groups_callback)
    return AuthTktAuthenticationPolicy(**kwargs)


def acl_factory(**kwargs):
    return ACLAuthorizationPolicy()


cfg.yafowil = Properties()
cfg.yafowil.js_skip = set()
cfg.yafowil.css_skip = set()

# ignore bootstrap dependencies delivered by yafowil.bootstrap
cfg.yafowil.js_skip.add('bootstrap.dependencies')
cfg.yafowil.css_skip.add('bootstrap.dependencies')


class YafowilResources(YafowilResourcesBase):

    def __init__(self, js_skip=[], css_skip=[], config=None):
        self.config = config
        super(YafowilResources, self).__init__(
            js_skip=js_skip,
            css_skip=css_skip
        )

    def configure_resource_directory(self, plugin_name, resourc_edir):
        app = sys.modules[__name__]
        resources_view = static_view(resourc_edir, use_subpath=True)
        view_name = '%s_resources' % plugin_name.replace('.', '_')
        setattr(app, view_name, resources_view)
        view_path = 'cone.app.%s' % view_name
        resource_base = '++resource++%s' % plugin_name
        self.config.add_view(view_path, name=resource_base)
        return resource_base


def configure_yafowil_addon_resources(config, public):
    resources = YafowilResources(
        js_skip=cfg.yafowil.js_skip,
        css_skip=cfg.yafowil.css_skip,
        config=config
    )
    js_resources = cfg.js.public if public else cfg.js.protected
    css_resources = cfg.css.public if public else cfg.css.protected
    for js in reversed(resources.js_resources):
        js_resources.insert(0, js)
    for css in resources.css_resources:
        css_resources.insert(0, css)


def configure_bdajax_resources():
    # bdajax needs to be loaded before resources depending on it order to
    # avoid double binding on document ready
    cfg.js.public.insert(0, '++resource++bdajax/bdajax_bs3.js')
    cfg.js.public.insert(0, '++resource++bdajax/bdajax.js')
    cfg.js.public.insert(0, '++resource++bdajax/overlay.js')


@adapter(IApplicationNode)
class ApplicationNodeTraverser(ResourceTreeTraverser):

    def __call__(self, request):
        result = super(ApplicationNodeTraverser, self).__call__(request)
        context = result['context']
        if not IApplicationNode.providedBy(context):
            if INode.providedBy(context):
                result['context'] = context.acquire(IApplicationNode)
                result['view_name'] = context.name
                result['traversed'] = tuple(node_path(result['context']))
            else:
                result['context'] = get_root()
                result['view_name'] = ''
                result['traversed'] = tuple()
        return result


def start_thread_monitor():  # pragma: no cover
    if not thread_shutdown_hooks:
        return

    def _monitor():
        main_thread = threading.main_thread()
        main_thread.join()
        for hook in thread_shutdown_hooks:
            hook()

    monitor = threading.Thread(target=_monitor)
    monitor.daemon = True
    monitor.start()


def main(global_config, **settings):
    """Returns WSGI application.
    """
    # set authentication related application properties
    security.ADMIN_USER = settings.get('cone.admin_user')
    security.ADMIN_PASSWORD = settings.get('cone.admin_password')
    security.AUTHENTICATOR = settings.get('cone.authenticator')

    auth_secret = settings.pop('cone.auth_secret', 'secret')
    auth_cookie_name = settings.pop('cone.auth_cookie_name', 'auth_tkt')
    auth_secure = settings.pop('cone.auth_secure', False)
    auth_include_ip = settings.pop('cone.auth_include_ip', False)
    auth_timeout = settings.pop('cone.auth_timeout', None)
    auth_reissue_time = settings.pop('cone.auth_reissue_time', None)
    if auth_reissue_time is not None:
        auth_reissue_time = int(auth_reissue_time)
    auth_max_age = settings.pop('cone.auth_max_age', None)
    if auth_max_age is not None:
        auth_max_age = int(auth_max_age)
    auth_http_only = settings.pop('cone.auth_http_only', False)
    auth_path = settings.pop('cone.auth_path', "/")
    auth_wild_domain = settings.pop('cone.auth_wild_domain', True)

    auth_policy = auth_tkt_factory(
        secret=auth_secret,
        cookie_name=auth_cookie_name,
        secure=auth_secure,
        include_ip=auth_include_ip,
        timeout=auth_timeout,
        reissue_time=auth_reissue_time,
        max_age=auth_max_age,
        http_only=auth_http_only,
        path=auth_path,
        wild_domain=auth_wild_domain,
    )

    # create root node
    global root
    root_node_factory = settings.pop('cone.root.node_factory', None)
    if root_node_factory:
        root = import_from_string(root_node_factory)(settings)
    else:
        root = default_root_node_factory(settings)

    if settings.get('testing.hook_global_registry'):
        globalreg = getGlobalSiteManager()
        config = Configurator(registry=globalreg)
        config.setup_registry(root_factory=get_root, settings=settings)
    else:
        config = Configurator(root_factory=get_root, settings=settings)

    # set authentication and authorization policies
    config.set_authentication_policy(auth_policy)
    config.set_authorization_policy(acl_factory())
    config.commit()

    # begin configuration
    config.begin()

    # include general dependencies
    config.include(pyramid_chameleon)
    config.include(pyramid_zcml)

    # add custom traverser
    config.registry.registerAdapter(ApplicationNodeTraverser)

    # available languages
    available_languages = settings.get('cone.available_languages', '')
    cfg.available_languages = [
        lang.strip() for lang in available_languages.split(',') if lang
    ]

    # main template
    main_template = settings.get('cone.main_template')
    if main_template:
        cfg.main_template = main_template

    # add translation
    config.add_translation_dirs('cone.app:locale/')

    # XXX: register yafowil and all yafowil addon widget locales.
    #      provide locales either in yafowil resources or as entry points in
    #      all yafowil packages providing translations

    # static routes
    config.add_route("favicon", "/favicon.ico")
    # XXX: robots.txt
    # XXX: humans.txt

    # register static resources
    config.add_view(browser.static_resources, name='static')

    # scan browser package
    config.scan(browser)

    # load zcml
    config.load_zcml('configure.zcml')

    # read plugin configurator
    plugins = settings.get('cone.plugins', '')
    plugins = plugins.split('\n')
    plugins = [pl for pl in plugins if pl and not pl.startswith('#')]
    for plugin in plugins:
        try:
            importlib.import_module(plugin)
        except ImportError:
            msg = 'Cannot import plugin {}\n{}'.format(
                plugin,
                format_traceback()
            )
            logger.error(msg)
            continue
        try:
            config.load_zcml('{}:configure.zcml'.format(plugin))
        except IOError:  # pragma: no cover
            msg = 'No configure.zcml in {}'.format(plugin)
            logger.info(msg)

    # execute main hooks
    filtered_hooks = list()
    for plugin in plugins:
        for hook in main_hooks:
            if hook.__module__.startswith(plugin):
                filtered_hooks.append(hook)
    for hook in filtered_hooks:
        hook(config, global_config, settings)

    # load and initialize UGM
    backend_name = settings.get('ugm.backend')
    # B/C
    if not backend_name:
        backend_name = settings.get('cone.auth_impl')
    if backend_name:
        try:
            ugm_backend.load(backend_name, settings)
            ugm_backend.initialize()
        except Exception:  # pragma: no cover
            msg = 'Failed to create UGM backend:\n{}'.format(format_traceback())
            logger.error(msg)

    user_display_attr = settings.get('ugm.user_display_attr')
    if user_display_attr:
        ugm_backend.user_display_attr = user_display_attr

    group_display_attr = settings.get('ugm.group_display_attr')
    if group_display_attr:
        ugm_backend.group_display_attr = group_display_attr

    # register yafowil static resources
    # done after addon config - addon code may disable yafowil resource groups
    # XXX: ``yafowil.resources_public`` is a temporary hack and stays
    # undocumented. In 1.1. ``webresource`` will be used for resource
    # registration and resource delivery configuration.
    yafowil_resources_public = settings.get('yafowil.resources_public')
    yafowil_resources_public = yafowil_resources_public in ['1', 'True', 'true']
    configure_yafowil_addon_resources(config, yafowil_resources_public)

    # ensure bdajax resources gets loaded before resources depending on it
    configure_bdajax_resources()

    # end configuration
    config.end()

    # start thread monitor if thread shutdown hooks registered
    start_thread_monitor()

    # return wsgi app
    return config.make_wsgi_app()


def make_remote_addr_middleware(app, global_conf):
    return RemoteAddrFilter(app)


class RemoteAddrFilter(object):
    """Use this middleware if nginx is used as proxy and IP address should be
    included in auth cookie. make sure nginx passes the right header:

    proxy_set_header X-Real-IP $remote_addr;
    """

    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        if 'HTTP_X_REAL_IP' in environ:
            environ['REMOTE_ADDR'] = environ['HTTP_X_REAL_IP']
        return self.app(environ, start_response)



from cone.app import get_root
from cone.app import testing
from cone.app.browser import favicon_view
from cone.app.browser import get_related_view
from cone.app.browser import main_view
from cone.app.browser import RelatedViewConsumer
from cone.app.browser import RelatedViewProvider
from cone.app.browser import render_main_template
from cone.app.browser import set_related_view
from cone.app.browser.utils import make_url
from cone.app.model import BaseNode
from cone.tile import Tile
from cone.tile.tests import TileTestCase
from plumber import plumbing
from webob.response import Response


class BrowserTest(TileTestCase):
    layer = testing.security

    def test_render_main_template(self):
        root = get_root()
        request = self.layer.new_request()
        response = render_main_template(root, request)
        self.assertTrue(isinstance(response, Response))

    def test_main_view(self):
        root = get_root()
        request = self.layer.new_request()
        response = main_view(root, request)
        self.assertTrue(isinstance(response, Response))

    def test_favicon_view(self):
        request = self.layer.new_request()
        response = favicon_view(request)
        self.assertTrue(isinstance(response, Response))
        self.assertEqual(
            response.headers['Content-Type'],
            'image/vnd.microsoft.icon'
        )

    def test_related_view_support(self):
        # Test ``set_related_view``
        request = self.layer.new_request()
        set_related_view(request, 'someview')
        self.assertEqual(
            request.environ['cone.app.related_view'],
            'someview'
        )

        # Test ``get_related_view``
        self.assertEqual(get_related_view(request), 'someview')

        # Test ``RelatedViewConsumer``
        @plumbing(RelatedViewConsumer)
        class RelatedViewConsumingTile(Tile):
            def render(self):
                return make_url(
                    self.request,
                    node=self.model,
                    resource=self.related_view)

        model = BaseNode(name='root')
        request = self.layer.new_request()
        tile = RelatedViewConsumingTile()
        self.assertEqual(
            tile(model, request),
            'http://example.com/root'
        )

        set_related_view(request, 'someview')
        self.assertEqual(
            tile(model, request),
            'http://example.com/root/someview'
        )

        # Test ``RelatedViewProvider``
        @plumbing(RelatedViewProvider)
        class RelatedViewProvidingTile(Tile):
            related_view = 'related_view'

            def render(self):
                return RelatedViewConsumingTile()(self.model, self.request)

        model = BaseNode(name='root')
        request = self.layer.new_request()
        tile = RelatedViewProvidingTile()
        self.assertEqual(
            tile(model, request),
            'http://example.com/root/related_view'
        )



from cone.app import compat
from cone.app import testing
from cone.app.browser.ajax import AjaxEvent
from cone.app.browser.ajax import AjaxMessage
from cone.app.browser.ajax import AjaxPath
from cone.app.browser.authoring import _FormRenderingTile
from cone.app.browser.authoring import add
from cone.app.browser.authoring import AddFormHeading
from cone.app.browser.authoring import CameFromNext
from cone.app.browser.authoring import ContentAddForm
from cone.app.browser.authoring import ContentEditForm
from cone.app.browser.authoring import ContentForm
from cone.app.browser.authoring import edit
from cone.app.browser.authoring import EditFormHeading
from cone.app.browser.authoring import FormHeading
from cone.app.browser.authoring import is_ajax
from cone.app.browser.authoring import overlayadd
from cone.app.browser.authoring import OverlayAddForm
from cone.app.browser.authoring import overlayedit
from cone.app.browser.authoring import OverlayEditForm
from cone.app.browser.authoring import OverlayForm
from cone.app.browser.authoring import overlayform
from cone.app.browser.authoring import render_form
from cone.app.browser.form import Form
from cone.app.model import AdapterNode
from cone.app.model import BaseNode
from cone.app.model import get_node_info
from cone.app.model import node_info
from cone.app.model import NodeInfo
from cone.app.model import register_node_info
from cone.tile import render_tile
from cone.tile import tile
from cone.tile.tests import TileTestCase
from plumber import plumbing
from webob.exc import HTTPFound
from yafowil.base import factory
from zope.interface import implementer
from zope.interface import Interface


class TestBrowserAuthoring(TileTestCase):
    layer = testing.security

    def test_is_ajax(self):
        # Helper for detecting ajax forms
        request = self.layer.new_request()
        self.assertFalse(is_ajax(request))

        request.params['ajax'] = '1'
        self.assertTrue(is_ajax(request))

    def test_render_form(self):
        # Form rendering helper
        with self.layer.hook_tile_reg():
            @tile(name='someform', permission='login')
            class SomeForm(Form):
                def prepare(self):
                    self.form = form = factory('form', name='someform')
                    form['somefield'] = factory(
                        'text',
                        props={
                            'label': 'Field'
                        })

        # Regular page view, render 'someform' tile as content area in main
        # template
        model = BaseNode()
        request = self.layer.new_request()
        res = render_form(model, request, 'someform')

        self.assertTrue(res.text.find('<!DOCTYPE html>') > -1)
        self.assertTrue(res.text.find('<form class="ajax"') > -1)
        self.assertTrue(res.text.find('id="form-someform"') > -1)

        # Ajax flag set, form has been submitted to hidden iframe. Form gets
        # rendered to dedicated container and hooked up to parent frame in
        # browser via some JS
        request.params['ajax'] = '1'
        res = render_form(model, request, 'someform')

        self.assertTrue(res.text.find('<div id="ajaxform">') > -1)
        self.assertTrue(res.text.find('parent.bdajax.render_ajax_form') > -1)

        # Form rendering tile. Has been introduced to handle node information
        # in add forms and is used in overlay and edit forms as mixin as well.
        # Simply renders another tile as form on ``render``.
        # XXX: Feels superfluous. Should be refactored and removed somwhen
        with self.layer.hook_tile_reg():
            @tile(name='someformrenderingtile', permission='login')
            class SomeFormTileRenderingTile(_FormRenderingTile):
                form_tile_name = 'someform'

        request = self.layer.new_request()
        self.checkOutput("""
        <form class="ajax" ... id="form-someform" ...</form>
        """, render_tile(model, request, 'someformrenderingtile'))

    def test_CameFromNext(self):
        # Plumbing behavior to hook up redirection after successful form
        # processing
        with self.layer.hook_tile_reg():
            @tile(name='camefromnextform')
            @plumbing(CameFromNext)
            class CameFromNextForm(Form):
                def prepare(self):
                    form = factory(
                        u'form',
                        name='camefromnextform',
                        props={'action': self.nodeurl})
                    form['next'] = factory(
                        'submit',
                        props={
                            'action': 'next',
                            'expression': True,
                            'next': self.next,
                            'label': 'Next',
                        })
                    self.form = form

        # Check behavior config defaults
        self.assertTrue(CameFromNextForm.default_came_from is None)
        self.assertTrue(CameFromNextForm.write_history_on_next is False)

        # Create a test model and login
        root = BaseNode()
        model = root['child'] = BaseNode()

        # Check whether ``came_from`` is rendered on form as proxy field
        with self.layer.authenticated('manager'):
            request = self.layer.new_request()
            came_from = compat.quote('http://example.com/some/path?foo=bar')
            request.params['came_from'] = came_from
            res = render_tile(model, request, 'camefromnextform')

        self.checkOutput("""
        ...<input id="input-camefromnextform-came_from"
        name="came_from" type="hidden"
        value="http%3A//example.com/some/path%3Ffoo%3Dbar" />...
        """, res)

        # No ``came_from`` on request, no ``default_came_from``, no ajax request
        with self.layer.authenticated('manager'):
            request = self.layer.new_request()
            request.params['action.camefromnextform.next'] = '1'
            res = render_tile(model, request, 'camefromnextform')

        self.assertTrue(isinstance(request.environ['redirect'], HTTPFound))
        self.assertEqual(
            request.environ['redirect'].location,
            'http://example.com/child'
        )

        # No ``came_from`` on request, ``default_came_from`` set to ``parent``,
        # no ajax request
        with self.layer.authenticated('manager'):
            CameFromNextForm.default_came_from = 'parent'
            res = render_tile(model, request, 'camefromnextform')

        self.assertTrue(isinstance(request.environ['redirect'], HTTPFound))
        self.assertEqual(
            request.environ['redirect'].location,
            'http://example.com/'
        )

        # No ``came_from`` on request, ``default_came_from`` set to URL, no
        # ajax request
        with self.layer.authenticated('manager'):
            came_from = compat.quote('http://example.com/foo/bar?baz=1')
            CameFromNextForm.default_came_from = came_from
            res = render_tile(model, request, 'camefromnextform')

        self.assertTrue(isinstance(request.environ['redirect'], HTTPFound))
        self.assertEqual(
            request.environ['redirect'].location,
            'http://example.com/foo/bar?baz=1'
        )

        # No ``came_from`` on request, ``default_came_from`` set to wrong
        # domain, no ajax request
        with self.layer.authenticated('manager'):
            CameFromNextForm.default_came_from = 'http://other.com'
            res = render_tile(model, request, 'camefromnextform')

        self.assertTrue(isinstance(request.environ['redirect'], HTTPFound))
        self.assertEqual(
            request.environ['redirect'].location,
            'http://example.com/child'
        )

        # ``came_from`` set to empty value on request, overrules
        # ``default_came_from``, no ajax request
        with self.layer.authenticated('manager'):
            CameFromNextForm.default_came_from = 'parent'
            request.params['came_from'] = ''
            res = render_tile(model, request, 'camefromnextform')

        self.assertTrue(isinstance(request.environ['redirect'], HTTPFound))
        self.assertEqual(
            request.environ['redirect'].location,
            'http://example.com/child'
        )

        # ``came_from`` set to ``parent`` on request, overrules
        # ``default_came_from``, no ajax request
        with self.layer.authenticated('manager'):
            CameFromNextForm.default_came_from = None
            request.params['came_from'] = 'parent'
            res = render_tile(model, request, 'camefromnextform')

        self.assertTrue(isinstance(request.environ['redirect'], HTTPFound))
        self.assertEqual(
            request.environ['redirect'].location,
            'http://example.com/'
        )

        # ``came_from`` set to URL on request, overrules ``default_came_from``,
        # no ajax request
        with self.layer.authenticated('manager'):
            came_from = compat.quote('http://example.com/default')
            CameFromNextForm.default_came_from = came_from

            came_from = compat.quote('http://example.com/other')
            request.params['came_from'] = came_from
            res = render_tile(model, request, 'camefromnextform')

        self.assertTrue(isinstance(request.environ['redirect'], HTTPFound))
        self.assertEqual(
            request.environ['redirect'].location,
            'http://example.com/other'
        )

        # Reset ``default_came_from``
        CameFromNextForm.default_came_from = None

        # ``came_from`` set to empty value on request, ajax request, no ajax
        # path continuation
        with self.layer.authenticated('manager'):
            request = self.layer.new_request()
            request.params['ajax'] = '1'
            request.params['action.camefromnextform.next'] = '1'
            request.params['came_from'] = ''
            res = render_tile(model, request, 'camefromnextform')

        self.assertTrue(len(request.environ['cone.app.continuation']) == 1)
        continuation = request.environ['cone.app.continuation'][0]
        self.assertTrue(isinstance(continuation, AjaxEvent))
        self.assertEqual(
            (continuation.target, continuation.name, continuation.selector),
            ('http://example.com/child', 'contextchanged', '#layout')
        )

        # ``came_from`` set to ``parent`` on request, ajax request, no ajax
        # path continuation
        with self.layer.authenticated('manager'):
            request.params['came_from'] = 'parent'
            res = render_tile(model, request, 'camefromnextform')

        self.assertTrue(len(request.environ['cone.app.continuation']) == 1)
        continuation = request.environ['cone.app.continuation'][0]
        self.assertTrue(isinstance(continuation, AjaxEvent))
        self.assertEqual(
            (continuation.target, continuation.name, continuation.selector),
            ('http://example.com/', 'contextchanged', '#layout')
        )

        # ``came_from`` set to URL on request, ajax request, no ajax path
        # continuation
        with self.layer.authenticated('manager'):
            came_from = compat.quote('http://example.com/some/path?foo=bar')
            request.params['came_from'] = came_from
            res = render_tile(model, request, 'camefromnextform')

        self.assertTrue(len(request.environ['cone.app.continuation']) == 1)
        continuation = request.environ['cone.app.continuation'][0]
        self.assertTrue(isinstance(continuation, AjaxEvent))
        self.assertEqual(
            (continuation.target, continuation.name, continuation.selector),
            ('http://example.com/some/path?foo=bar', 'contextchanged', '#layout')
        )

        # ``came_from`` set to wrong domain on request, ajax request, no ajax
        # path continuation
        with self.layer.authenticated('manager'):
            came_from = compat.quote('http://other.com')
            request.params['came_from'] = came_from
            res = render_tile(model, request, 'camefromnextform')

        self.assertTrue(len(request.environ['cone.app.continuation']) == 1)
        continuation = request.environ['cone.app.continuation'][0]
        self.assertTrue(isinstance(continuation, AjaxEvent))
        self.assertEqual(
            (continuation.target, continuation.name, continuation.selector),
            ('http://example.com/child', 'contextchanged', '#layout')
        )

        # ``came_from`` set to empty value on request, ajax request, setting
        # browser history configured
        with self.layer.authenticated('manager'):
            CameFromNextForm.write_history_on_next = True
            request = self.layer.new_request()
            request.params['ajax'] = '1'
            request.params['action.camefromnextform.next'] = '1'
            request.params['came_from'] = ''
            res = render_tile(model, request, 'camefromnextform')

        self.assertTrue(len(request.environ['cone.app.continuation']) == 2)

        path = request.environ['cone.app.continuation'][0]
        self.assertTrue(isinstance(path, AjaxPath))
        self.assertEqual(
            (path.path, path.target, path.event),
            (u'child', 'http://example.com/child', 'contextchanged:#layout')
        )

        event = request.environ['cone.app.continuation'][1]
        self.assertTrue(isinstance(event, AjaxEvent))
        self.assertEqual(
            (event.target, continuation.name, continuation.selector),
            ('http://example.com/child', 'contextchanged', '#layout')
        )

        # ``came_from`` set to ``parent`` on request, ajax request, setting
        # browser history configured
        with self.layer.authenticated('manager'):
            request.params['came_from'] = 'parent'
            res = render_tile(model, request, 'camefromnextform')

        self.assertTrue(len(request.environ['cone.app.continuation']) == 2)

        path = request.environ['cone.app.continuation'][0]
        self.assertTrue(isinstance(path, AjaxPath))
        self.assertEqual(
            (path.path, path.target, path.event),
            ('', 'http://example.com/', 'contextchanged:#layout')
        )

        event = request.environ['cone.app.continuation'][1]
        self.assertTrue(isinstance(event, AjaxEvent))
        self.assertEqual(
            (event.target, continuation.name, continuation.selector),
            ('http://example.com/', 'contextchanged', '#layout')
        )

        # ``came_from`` set to URL on request, ajax request, setting browser
        # history configured
        with self.layer.authenticated('manager'):
            came_from = compat.quote('http://example.com/some/path')
            request.params['came_from'] = came_from
            res = render_tile(model, request, 'camefromnextform')

        self.assertTrue(len(request.environ['cone.app.continuation']) == 2)

        path = request.environ['cone.app.continuation'][0]
        self.assertTrue(isinstance(path, AjaxPath))
        self.assertEqual(
            (path.path, path.target, path.event),
            ('/some/path', 'http://example.com/some/path', 'contextchanged:#layout')
        )

        event = request.environ['cone.app.continuation'][1]
        self.assertTrue(isinstance(event, AjaxEvent))
        self.assertEqual(
            (event.target, continuation.name, continuation.selector),
            ('http://example.com/some/path', 'contextchanged', '#layout')
        )

        # ``came_from`` set to to wrong on request, ajax request, setting
        # browser history configured
        with self.layer.authenticated('manager'):
            came_from = compat.quote('http://other.com')
            request.params['came_from'] = came_from
            res = render_tile(model, request, 'camefromnextform')

        self.assertTrue(len(request.environ['cone.app.continuation']) == 2)

        path = request.environ['cone.app.continuation'][0]
        self.assertTrue(isinstance(path, AjaxPath))
        self.assertEqual(
            (path.path, path.target, path.event),
            (u'child', 'http://example.com/child', 'contextchanged:#layout')
        )

        event = request.environ['cone.app.continuation'][1]
        self.assertTrue(isinstance(event, AjaxEvent))
        self.assertEqual(
            (event.target, continuation.name, continuation.selector),
            ('http://example.com/child', 'contextchanged', '#layout')
        )

        # Reset ``write_history_on_next``
        CameFromNextForm.write_history_on_next = False

    def test_FormHeading(self):
        # Abstract form heading
        @plumbing(FormHeading)
        class FormWithHeading(object):
            pass

        form_with_heading = FormWithHeading()
        err = self.expectError(
            NotImplementedError,
            lambda: form_with_heading.form_heading
        )
        expected = 'Abstract ``FormHeading`` does not implement ``form_heading``'
        self.assertEqual(str(err), expected)

    def test_ContentForm(self):
        @plumbing(ContentForm)
        class MyForm(Form):
            def prepare(self):
                form = factory(
                    u'form',
                    name='myform',
                    props={
                        'action': self.nodeurl
                    })
                self.form = form

        model = BaseNode()
        request = self.layer.new_request()
        content_form = MyForm()
        content_form.model = model
        content_form.request = request

        self.assertTrue(content_form.show_heading)
        self.assertTrue(content_form.show_contextmenu)
        # content_form.form_heading is supposed to be overwritten
        self.assertEqual(content_form.form_heading, 'content_form_heading')

        with self.layer.authenticated('max'):
            res = content_form.rendered_contextmenu
        self.assertTrue(res.find('<div id="contextmenu"') > -1)

        with self.layer.authenticated('max'):
            res = content_form(model, request)
        expected = '<div class="panel-heading content-heading">'
        self.assertTrue(res.find(expected) > -1)

    @testing.reset_node_info_registry
    def test_AddFormHeading(self):
        @node_info(
            name='addnode',
            title='Add Node')
        class AddNode(BaseNode):
            pass

        @plumbing(AddFormHeading)
        class AddForm(Form):
            pass

        add_form = AddForm()
        add_form.model = AddNode()
        add_form.request = self.layer.new_request()
        self.assertEqual(add_form.form_heading, 'Add: Add Node')

    @testing.reset_node_info_registry
    def test_adding(self):
        # Provide a node interface needed for different node style binding to
        # test form
        class ITestAddingNode(Interface):
            pass

        # Create dummy node
        @node_info(
            name='mynode',
            title='My Node',
            description='This is My node.',
            addables=['mynode'])  # self containment
        @implementer(ITestAddingNode)
        class MyNode(BaseNode):
            pass

        # Create another dummy node inheriting from AdapterNode
        @node_info(
            name='myadapternode',
            title='My Adapter Node',
            description='This is My adapter node.',
            addables=['myadapternode'])  # self containment
        @implementer(ITestAddingNode)
        class MyAdapterNode(AdapterNode):
            pass

        # Create and register an ``addform`` named form tile
        with self.layer.hook_tile_reg():
            @tile(name='addform', interface=ITestAddingNode)
            @plumbing(ContentAddForm)
            class MyAddForm(Form):
                def prepare(self):
                    form = factory(
                        u'form',
                        name='addform',
                        props={
                            'action': self.nodeurl
                        })
                    form['id'] = factory(
                        'field:label:text',
                        props={
                            'label': 'Id',
                        })
                    form['title'] = factory(
                        'field:label:text',
                        props={
                            'label': 'Title',
                        })
                    form['add'] = factory(
                        'submit',
                        props={
                            'action': 'add',
                            'expression': True,
                            'handler': self.add,
                            'next': self.next,
                            'label': 'Add',
                        })
                    self.form = form

                def add(self, widget, data):
                    fetch = self.request.params.get
                    child = MyNode()
                    child.attrs.title = fetch('addform.title')
                    self.model.parent[fetch('addform.id')] = child
                    self.model = child

        # Create dummy container
        root = MyNode()

        # Render without factory
        with self.layer.authenticated('manager'):
            request = self.layer.new_request()
            self.assertEqual(
                render_tile(root, request, 'add'),
                u'unknown_factory'
            )

        # Render with valid factory
        with self.layer.authenticated('manager'):
            request.params['factory'] = 'mynode'
            result = render_tile(root, request, 'add')

        self.assertTrue(result.find(u'<form action="http://example.com"') != -1)

        # Render with valid factory on adapter node
        with self.layer.authenticated('manager'):
            adapterroot = MyAdapterNode(None, None, None)
            request.params['factory'] = 'myadapternode'
            result = render_tile(adapterroot, request, 'add')

        self.assertTrue(result.find(u'<form action="http://example.com"') != -1)

        # Render with submitted data
        with self.layer.authenticated('manager'):
            request = self.layer.current_request
            request.params['factory'] = 'mynode'
            request.params['action.addform.add'] = '1'
            request.params['addform.id'] = 'somechild'
            request.params['addform.title'] = 'Some Child'
            render_tile(root, request, 'add')

        self.assertTrue(isinstance(request.environ['redirect'], HTTPFound))
        self.checkOutput("""
        <class '...MyNode'>: None
          <class '...MyNode'>: somechild
        """, root.treerepr())

        self.assertEqual(
            request.environ['redirect'].location,
            'http://example.com/somechild'
        )
        del request.environ['redirect']

        # Render with 'came_from' set
        with self.layer.authenticated('manager'):
            request.params['came_from'] = 'parent'
            render_tile(root, request, 'add')

        self.assertEqual(
            request.environ['redirect'].location,
            'http://example.com/'
        )
        del request.environ['redirect']

        with self.layer.authenticated('manager'):
            came_from = compat.quote('http://example.com/foo/bar?baz=1')
            request.params['came_from'] = came_from
            render_tile(root, request, 'add')

        self.assertEqual(
            request.environ['redirect'].location,
            'http://example.com/foo/bar?baz=1'
        )

        # Render with ajax flag
        with self.layer.authenticated('manager'):
            request.params['ajax'] = '1'
            render_tile(root, request, 'add')

        self.assertTrue(isinstance(
            request.environ['cone.app.continuation'][0],
            AjaxEvent
        ))

        # Check the modified model
        self.assertEqual(root.keys(), ['somechild'])
        self.assertEqual(root['somechild'].attrs.title, 'Some Child')

        # Add view
        with self.layer.authenticated('manager'):
            request = self.layer.new_request()
            request.params['factory'] = 'mynode'
            request.params['action.addform.add'] = '1'
            request.params['addform.id'] = 'somechild'
            request.params['addform.title'] = 'Some Child'
            res = add(root, request)

        self.assertTrue(isinstance(res, HTTPFound))

        with self.layer.authenticated('manager'):
            request.params['ajax'] = '1'
            res = str(add(root, request))

        self.assertTrue(res.find('parent.bdajax.render_ajax_form') != -1)

    @testing.reset_node_info_registry
    def test_EditFormHeading(self):
        @plumbing(EditFormHeading)
        class EditForm(Form):
            pass

        self.assertEqual(BaseNode().node_info_name, '')
        self.assertEqual(get_node_info(''), None)

        edit_form = EditForm()
        edit_form.model = BaseNode()
        edit_form.request = self.layer.new_request()
        self.assertEqual(edit_form.form_heading, 'edit')

        @node_info(
            name='editnode',
            title='Edit Node')
        class EditNode(BaseNode):
            pass

        edit_form = EditForm()
        edit_form.model = EditNode()
        edit_form.request = self.layer.new_request()
        self.assertEqual(edit_form.form_heading, 'Edit: Edit Node')

    @testing.reset_node_info_registry
    def test_editing(self):
        @node_info(
            name='mynode',
            title='My Node')
        class MyNode(BaseNode):
            pass

        # Create and register an ``editform`` named form tile
        with self.layer.hook_tile_reg():
            @tile(name='editform', interface=MyNode)
            @plumbing(ContentEditForm)
            class MyEditForm(Form):
                def prepare(self):
                    form = factory(
                        u'form',
                        name='editform',
                        props={
                            'action': self.nodeurl
                        })
                    form['title'] = factory(
                        'field:label:text',
                        value=self.model.attrs.title,
                        props={
                            'label': 'Title',
                        })
                    form['update'] = factory(
                        'submit',
                        props={
                            'action': 'update',
                            'expression': True,
                            'handler': self.update,
                            'next': self.next,
                            'label': 'Update',
                        })
                    self.form = form

                def update(self, widget, data):
                    fetch = self.request.params.get
                    self.model.attrs.title = fetch('editform.title')

        # Dummy model
        root = MyNode()
        child = root['somechild'] = MyNode()
        child.attrs.title = 'My Node'

        # Render form with value from model
        with self.layer.authenticated('editor'):
            request = self.layer.new_request()
            res = render_tile(root['somechild'], request, 'edit')

        self.checkOutput("""
        ...<span class="label label-primary">Edit: My Node</span>...
        <form action="http://example.com/somechild"...
        """, res)

        # Render with submitted data. Default next URL of EditForm is the
        # edited node
        with self.layer.authenticated('editor'):
            request = self.layer.new_request()
            request.params['action.editform.update'] = '1'
            request.params['editform.title'] = 'Changed title'
            res = render_tile(root['somechild'], request, 'edit')

        self.assertEqual(
            request.environ['redirect'].location,
            'http://example.com/somechild'
        )

        # Check next URL with ``parent`` as ``came_from`` value
        with self.layer.authenticated('editor'):
            request = self.layer.new_request()
            request.params['action.editform.update'] = '1'
            request.params['editform.title'] = 'Changed title'
            request.params['came_from'] = 'parent'
            res = render_tile(root['somechild'], request, 'edit')

        self.assertEqual(
            request.environ['redirect'].location,
            'http://example.com/'
        )

        # Check next URL with URL as ``came_from`` value
        with self.layer.authenticated('editor'):
            request = self.layer.new_request()
            request.params['action.editform.update'] = '1'
            request.params['editform.title'] = 'Changed title'
            came_from = compat.quote('http://example.com/other/node/in/tree')
            request.params['came_from'] = came_from
            res = render_tile(root['somechild'], request, 'edit')

        self.assertEqual(
            request.environ['redirect'].location,
            'http://example.com/other/node/in/tree'
        )

        # Render with ajax flag
        with self.layer.authenticated('editor'):
            request = self.layer.new_request()
            request.params['action.editform.update'] = '1'
            request.params['editform.title'] = 'Changed title'
            request.params['ajax'] = '1'
            res = render_tile(root['somechild'], request, 'edit')

        self.assertTrue(isinstance(
            request.environ['cone.app.continuation'][0],
            AjaxEvent
        ))

        # URL computing is the same as if ``HTTPFound`` instance is returned.
        # In Ajax case, the URL is used as ajax target
        self.assertEqual(
            request.environ['cone.app.continuation'][0].target,
            'http://example.com/somechild'
        )

        with self.layer.authenticated('editor'):
            request = self.layer.new_request()
            request.params['action.editform.update'] = '1'
            request.params['editform.title'] = 'Changed title'
            came_from = compat.quote('http://example.com/other/node/in/tree')
            request.params['came_from'] = came_from
            request.params['ajax'] = '1'
            res = render_tile(root['somechild'], request, 'edit')

        self.assertEqual(
            request.environ['cone.app.continuation'][0].target,
            'http://example.com/other/node/in/tree'
        )
        # Check the updated node
        self.assertEqual(root['somechild'].attrs.title, 'Changed title')

        # Edit view
        with self.layer.authenticated('editor'):
            request = self.layer.new_request()
            request.params['action.editform.update'] = '1'
            request.params['editform.title'] = 'Changed title'
            root.attrs.title = 'Foo'
            res = edit(root, request)

        self.assertTrue(isinstance(res, HTTPFound))

        with self.layer.authenticated('editor'):
            request = self.layer.new_request()
            request.params['action.editform.update'] = '1'
            request.params['editform.title'] = 'Changed title'
            request.params['ajax'] = '1'
            res = str(edit(root, request))

        self.assertTrue(res.find('parent.bdajax.render_ajax_form') != -1)

    def test_deleting(self):
        class CallableNode(BaseNode):
            def __call__(self):
                pass

        node = CallableNode()
        node['child'] = CallableNode()
        self.checkOutput("""
        <class '...CallableNode'>: None
          <class '...CallableNode'>: child
        """, node.treerepr())

        del node['child']
        self.checkOutput("""
        <class '...CallableNode'>: None
        """, node.treerepr())

        node['child'] = CallableNode()

        with self.layer.authenticated('manager'):
            request = self.layer.new_request()
            self.assertEqual(render_tile(node['child'], request, 'delete'), u'')

        self.assertEqual(
            request.environ['cone.app.continuation'][0].payload,
            u'Object "child" not deletable'
        )

        node['child'].properties.action_delete = True

        with self.layer.authenticated('manager'):
            request = self.layer.new_request()
            self.assertEqual(render_tile(node['child'], request, 'delete'), u'')

        self.assertTrue(isinstance(
            request.environ['cone.app.continuation'][0],
            AjaxEvent
        ))
        self.assertTrue(isinstance(
            request.environ['cone.app.continuation'][1],
            AjaxMessage
        ))
        self.checkOutput("""
        <class '...CallableNode'>: None
        """, node.treerepr())

    @testing.reset_node_info_registry
    def test_add_items_dropdown(self):
        @node_info(
            name='mynode',
            addables=['mynode'])
        class MyNode(BaseNode):
            pass

        # Dummy model
        root = MyNode()
        root['somechild'] = MyNode()
        # child.attrs.title = 'My Node'

        # Dropdown menu containing links to the addforms of allowed child nodes
        with self.layer.authenticated('manager'):
            request = self.layer.new_request()
            rendered = render_tile(root['somechild'], request, 'add_dropdown')

        # Non JS link to add form
        expected = 'href="http://example.com/somechild/add?factory=mynode"'
        self.assertTrue(rendered.find(expected) != -1)

        # Ajax target for add form
        expected = 'ajax:target="http://example.com/somechild?factory=mynode"'
        self.assertTrue(rendered.find(expected) != -1)

        # Ajax action rule for add form
        expected = 'ajax:action="add:#content:inner"'
        self.assertTrue(rendered.find(expected) != -1)

        # Allow another node type as child
        nodeinfo = NodeInfo()
        register_node_info('anothernode', nodeinfo)
        get_node_info('mynode').addables = ['mynode', 'anothernode']

        with self.layer.authenticated('manager'):
            request = self.layer.new_request()
            rendered = render_tile(root['somechild'], request, 'add_dropdown')

        # Non JS links to add form
        expected = 'href="http://example.com/somechild/add?factory=mynode"'
        self.assertTrue(rendered.find(expected) != -1)

        expected = 'href="http://example.com/somechild/add?factory=anothernode"'
        self.assertTrue(rendered.find(expected) != -1)

        # Ajax targets for add form
        expected = 'ajax:target="http://example.com/somechild?factory=mynode"'
        self.assertTrue(rendered.find(expected) != -1)

        expected = 'ajax:target="http://example.com/somechild?factory=anothernode"'
        self.assertTrue(rendered.find(expected) != -1)

        # Test node without addables, results in empty listing.
        # XXX: hide entire widget if no items
        @node_info(name='nochildaddingnode')
        class NoChildAddingNode(BaseNode):
            pass

        with self.layer.authenticated('manager'):
            request = self.layer.new_request()
            rendered = render_tile(NoChildAddingNode(), request, 'add_dropdown')

        self.checkOutput("""
        ...<li class="dropdown">
        <a href="#"
        class="dropdown-toggle"
        data-toggle="dropdown">
        <span>Add</span>
        <span class="caret"></span>
        </a>
        <ul class="dropdown-menu" role="addmenu">
        </ul>
        </li>...
        """, rendered)

        # Test node with invalid addable, results in empty listing
        # XXX: hide entire widget if no items
        @node_info(
            name='invalidchildnodeinfo',
            addables=['invalid'])
        class InvalidChildNodeInfoNode(BaseNode):
            pass

        with self.layer.authenticated('manager'):
            request = self.layer.new_request()
            rendered = render_tile(
                InvalidChildNodeInfoNode(),
                request,
                'add_dropdown'
            )
        self.checkOutput("""
        ...<li class="dropdown">
        <a href="#"
        class="dropdown-toggle"
        data-toggle="dropdown">
        <span>Add</span>
        <span class="caret"></span>
        </a>
        <ul class="dropdown-menu" role="addmenu">
        </ul>
        </li>...
        """, rendered)

    def test_overlay_form(self):
        with self.layer.hook_tile_reg():
            @tile(name='overlayform', interface=BaseNode)
            @plumbing(OverlayForm)
            class MyOverlayForm(Form):
                def prepare(self):
                    form = factory(
                        u'form',
                        name='overlayform',
                        props={
                            'action': self.nodeurl + '/' + self.action_resource
                        })
                    form['title'] = factory(
                        'field:label:error:text',
                        value=self.model.attrs.title,
                        props={
                            'label': 'Title',
                            'required': 'Title is required'
                        })
                    form['update'] = factory(
                        'submit',
                        props={
                            'action': 'update',
                            'expression': True,
                            'handler': self.update,
                            'next': self.next,
                            'label': 'Update',
                        })
                    self.form = form

                def update(self, widget, data):
                    fetch = self.request.params.get
                    self.model.attrs.title = fetch('editform.title')

        model = BaseNode(name='root')
        model.attrs.title = u'Title'

        # Overlay form invocation happens via overlay form entry tile
        request = self.layer.new_request()
        request.params['ajax'] = '1'
        with self.layer.authenticated('max'):
            res = render_tile(model, request, 'overlayformtile')

        expected = '<form action="http://example.com/root/overlayform"'
        self.assertTrue(res.startswith(expected))
        expected = 'class="ajax"'
        self.assertTrue(res.find(expected) > -1)
        self.assertEqual(
            request.environ['cone.app.form.selector'],
            '#ajax-overlay .overlay_content'
        )
        self.assertEqual(request.environ['cone.app.form.mode'], 'inner')

        # Overlay form sumbmission happens via related pyramid view
        # Case form error
        request = self.layer.new_request()
        request.params['ajax'] = '1'
        request.params['overlayform.title'] = ''
        request.params['action.overlayform.update'] = '1'
        with self.layer.authenticated('max'):
            res = overlayform(model, request)

        expected = '<div id="ajaxform">'
        self.assertTrue(res.text.startswith(expected))
        expected = '<form action="http://example.com/root/overlayform"'
        self.assertTrue(res.text.find(expected) > -1)
        expected = '<div class="errormessage">Title is required</div>'
        self.assertTrue(res.text.find(expected) > -1)
        expected = '<script'
        self.assertTrue(res.text.find(expected) > -1)
        expected = (
            "parent.bdajax.render_ajax_form("
            "child, '#ajax-overlay .overlay_content', 'inner', false);"
        )
        self.assertTrue(res.text.find(expected) > -1)

        # Case form success
        request = self.layer.new_request()
        request.params['ajax'] = '1'
        request.params['overlayform.title'] = 'New Title'
        request.params['action.overlayform.update'] = '1'
        with self.layer.authenticated('max'):
            res = overlayform(model, request)

        expected = '<div id="ajaxform">'
        self.assertTrue(res.text.startswith(expected))
        self.assertFalse(res.text.find('<form') > -1)
        expected = '<script'
        self.assertTrue(res.text.find(expected) > -1)
        expected = (
            "parent.bdajax.render_ajax_form("
            "child, '#ajax-overlay .overlay_content', 'inner', ["
        )
        self.assertTrue(res.text.find(expected) > -1)
        expected = '"close": true'
        self.assertTrue(res.text.find(expected) > -1)

    @testing.reset_node_info_registry
    def test_overlay_add(self):
        @node_info(
            name='mynode',
            addables=['mynode'])
        class MyNode(BaseNode):
            pass

        with self.layer.hook_tile_reg():
            @tile(name='overlayaddform', interface=MyNode)
            @plumbing(OverlayAddForm)
            class MyOverlayAddForm(Form):
                def prepare(self):
                    form = factory(
                        u'form',
                        name='overlayaddform',
                        props={
                            'action': self.nodeurl + '/' + self.action_resource
                        })
                    form['title'] = factory(
                        'field:label:error:text',
                        props={
                            'label': 'Title',
                            'required': 'Title is required'
                        })
                    form['add'] = factory(
                        'submit',
                        props={
                            'action': 'add',
                            'expression': True,
                            'handler': self.add,
                            'next': self.next,
                            'label': 'Add',
                        })
                    self.form = form

                def add(self, widget, data):
                    fetch = self.request.params.get
                    child = MyNode()
                    child.attrs.title = fetch('overlayaddform.title')
                    self.model.parent['new'] = child
                    self.model = child

        root = MyNode(name='root')

        # Overlay addform invocation happens via entry tile
        request = self.layer.new_request()
        request.params['ajax'] = '1'
        request.params['factory'] = 'mynode'

        with self.layer.authenticated('max'):
            res = render_tile(root, request, 'overlayadd')

        expected = '<form action="http://example.com/root/overlayadd"'
        self.assertTrue(res.startswith(expected))
        expected = 'class="ajax"'
        self.assertTrue(res.find(expected) > -1)
        self.assertEqual(
            request.environ['cone.app.form.selector'],
            '#ajax-overlay .overlay_content'
        )
        self.assertEqual(request.environ['cone.app.form.mode'], 'inner')

        # Overlay addform sumbmission happens via related pyramid view
        # Case form error
        request = self.layer.new_request()
        request.params['ajax'] = '1'
        request.params['factory'] = 'mynode'
        request.params['overlayaddform.title'] = ''
        request.params['action.overlayaddform.add'] = '1'

        with self.layer.authenticated('max'):
            res = overlayadd(root, request)

        expected = '<div id="ajaxform">'
        self.assertTrue(res.text.startswith(expected))
        expected = '<form action="http://example.com/root/overlayadd"'
        self.assertTrue(res.text.find(expected) > -1)
        expected = '<div class="errormessage">Title is required</div>'
        self.assertTrue(res.text.find(expected) > -1)
        expected = '<script'
        self.assertTrue(res.text.find(expected) > -1)
        expected = (
            "parent.bdajax.render_ajax_form("
            "child, '#ajax-overlay .overlay_content', 'inner', false);"
        )
        self.assertTrue(res.text.find(expected) > -1)
        self.assertEqual(root.keys(), [])

        # Case form success
        request = self.layer.new_request()
        request.params['ajax'] = '1'
        request.params['factory'] = 'mynode'
        request.params['overlayaddform.title'] = 'Child'
        request.params['action.overlayaddform.add'] = '1'

        with self.layer.authenticated('max'):
            res = overlayadd(root, request)

        expected = '<div id="ajaxform">'
        self.assertTrue(res.text.startswith(expected))
        self.assertFalse(res.text.find('<form') > -1)
        expected = '<script'
        self.assertTrue(res.text.find(expected) > -1)
        expected = (
            "parent.bdajax.render_ajax_form("
            "child, '#ajax-overlay .overlay_content', 'inner', ["
        )
        self.assertTrue(res.text.find(expected) > -1)
        expected = '"close": true'
        self.assertTrue(res.text.find(expected) > -1)
        self.assertEqual(root.keys(), ['new'])
        self.assertEqual(root['new'].attrs.title, 'Child')

    @testing.reset_node_info_registry
    def test_overlay_edit(self):
        class MyNode(BaseNode):
            pass

        with self.layer.hook_tile_reg():
            @tile(name='overlayeditform', interface=MyNode)
            @plumbing(OverlayEditForm)
            class MyOverlayEditForm(Form):
                def prepare(self):
                    form = factory(
                        u'form',
                        name='overlayeditform',
                        props={
                            'action': self.nodeurl + '/' + self.action_resource
                        })
                    form['title'] = factory(
                        'field:label:error:text',
                        value=self.model.attrs.title,
                        props={
                            'label': 'Title',
                            'required': 'Title is required'
                        })
                    form['update'] = factory(
                        'submit',
                        props={
                            'action': 'update',
                            'expression': True,
                            'handler': self.update,
                            'next': self.next,
                            'label': 'Update',
                        })
                    self.form = form

                def update(self, widget, data):
                    fetch = self.request.params.get
                    self.model.attrs.title = fetch('overlayeditform.title')

        model = MyNode(name='model')
        model.attrs.title = 'My Title'

        # Overlay editform invocation happens via entry tile
        request = self.layer.new_request()
        request.params['ajax'] = '1'

        with self.layer.authenticated('max'):
            res = render_tile(model, request, 'overlayedit')

        expected = '<form action="http://example.com/model/overlayedit"'
        self.assertTrue(res.startswith(expected))
        expected = 'class="ajax"'
        self.assertTrue(res.find(expected) > -1)
        expected = 'value="My Title"'
        self.assertTrue(res.find(expected) > -1)
        self.assertEqual(
            request.environ['cone.app.form.selector'],
            '#ajax-overlay .overlay_content'
        )
        self.assertEqual(request.environ['cone.app.form.mode'], 'inner')

        # Overlay editform sumbmission happens via related pyramid view
        # Case form error
        request = self.layer.new_request()
        request.params['ajax'] = '1'
        request.params['overlayeditform.title'] = ''
        request.params['action.overlayeditform.update'] = '1'

        with self.layer.authenticated('max'):
            res = overlayedit(model, request)

        expected = '<div id="ajaxform">'
        self.assertTrue(res.text.startswith(expected))
        expected = '<form action="http://example.com/model/overlayedit"'
        self.assertTrue(res.text.find(expected) > -1)
        expected = '<div class="errormessage">Title is required</div>'
        self.assertTrue(res.text.find(expected) > -1)
        expected = '<script'
        self.assertTrue(res.text.find(expected) > -1)
        expected = (
            "parent.bdajax.render_ajax_form("
            "child, '#ajax-overlay .overlay_content', 'inner', false);"
        )
        self.assertTrue(res.text.find(expected) > -1)
        self.assertEqual(model.attrs.title, 'My Title')

        # Case form success
        request = self.layer.new_request()
        request.params['ajax'] = '1'
        request.params['overlayeditform.title'] = 'New Title'
        request.params['action.overlayeditform.update'] = '1'

        with self.layer.authenticated('max'):
            res = overlayedit(model, request)

        expected = '<div id="ajaxform">'
        self.assertTrue(res.text.startswith(expected))
        self.assertFalse(res.text.find('<form') > -1)
        expected = '<script'
        self.assertTrue(res.text.find(expected) > -1)
        expected = (
            "parent.bdajax.render_ajax_form("
            "child, '#ajax-overlay .overlay_content', 'inner', ["
        )
        self.assertTrue(res.text.find(expected) > -1)
        expected = '"close": true'
        self.assertTrue(res.text.find(expected) > -1)
        self.assertEqual(model.attrs.title, 'New Title')



from cone.app import compat
from datetime import datetime
from node.utils import safe_decode as _safe_decode
from node.utils import safe_encode as _safe_encode
import logging
import sys
import traceback


logger = logging.getLogger('cone.app')


def app_config():
    import cone.app
    return cone.app.cfg


# B/C
# deprecated: will be removed in cone.app 2.0
def safe_encode(value, encoding='utf-8'):
    return _safe_encode(value, encoding=encoding)


# B/C
# deprecated: will be removed in cone.app 2.0
def safe_decode(value, encoding='utf-8'):
    return _safe_decode(value, encoding=encoding)


def timestamp():
    return datetime.now()


def format_traceback():
    etype, value, tb = sys.exc_info()
    return ''.join(traceback.format_exception(etype, value, tb))


def node_path(node):
    # XXX: move to ``cone.app.utils``
    return [_safe_decode(p) for p in node.path if p is not None]


class DatetimeHelper(object):

    def w_value(self, val):
        if val is None:
            return 'None'
        if isinstance(val, datetime):
            return self.dt_to_iso(val)
        if isinstance(val, bytes):
            return val.decode('utf-8')
        return compat.UNICODE_TYPE(val)

    def r_value(self, val):
        try:
            return self.dt_from_iso(val)
        except (ValueError, TypeError):
            if isinstance(val, bytes):
                return val.decode('utf-8')
            return compat.UNICODE_TYPE(val)

    def dt_from_iso(self, str):
        return datetime.strptime(str, '%Y-%m-%dT%H:%M:%S')

    def dt_to_iso(self, dt):
        iso = dt.isoformat()
        if iso.find('.') != -1:
            iso = iso[:iso.rfind('.')]
        return iso


# XXX: move somewhere else, probably plumbing behavior for node
def add_creation_metadata(request, mapping):
    mapping['creator'] = request.authenticated_userid
    mapping['created'] = timestamp()
    mapping['modified'] = mapping['created']


# XXX: move somewhere else, probably plumbing behavior for node
def update_creation_metadata(request, mapping):
    mapping['modified'] = timestamp()
